"""
Resonance Driving Terms
-----------------------

This module contains RDT calculations related functionality of ``optics_measurements``.
It provides functions to compute global resonance driving terms **f_jklm**.
"""
from copy import deepcopy
from os.path import join

import numpy as np
import pandas as pd
import tfs
from scipy.optimize import curve_fit
from scipy.sparse import diags

from omc3.definitions.constants import PLANES
from omc3.optics_measurements.constants import ERR, EXT, AMPLITUDE
from omc3.optics_measurements.toolbox import df_diff
from omc3.utils import iotools, logging_tools, stats
from optics_functions.rdt import get_all_to_order, jklm2str

NBPMS_FOR_90 = 3
LOGGER = logging_tools.get_logger(__name__)

def _generate_plane_rdts(order):
    """
    This helper function generates two dictionnaries representing on what plane(s)
    a RDT can be seen and which tune it is a multiple of.

    For a given j+k+l+m = n multiple order, a line can be seen:
      - on the horizontal axis if j != 0, at:
          (1 - j + k) Qx + (m - l) Qy
      - on the vertical axis if l != 0, at:
          (k - j) Qx + (1 - l + m) Qy

    Reference equations are (3.27) and (3.28) in
    [Andrea Franchi's Thesis](https://repository.gsi.de/record/55413/files/GSI-Diss-2006-07.pdf)
    """
    # Get all the valid RDTs up to a certain order
    all_rdts = get_all_to_order(order)

    single_plane = {'X': [], 'Y': []}
    double_plane = {'X': [], 'Y': []}
    # Iterate through our RDTs and classify them depending on what plane they act
    for (j,k,l,m) in all_rdts:
        if j == 0 and l == 0:  # the RDT can't be seen on any plane
            continue
        if l+m == 0 and j != 0:  # The line where the RDT is seen is a multiple of the Qx line
            single_plane['X'].append((j,k,l,m))  # e.g. f1400, f3000, f4000
        elif j+k == 0 and l != 0:  # same, but for the Qy line
            single_plane['Y'].append((j,k,l,m))  # e.g. f0030,f0040
        elif j == 0:  # The RDT can only be seen on the vertical plane and uses both Qx and Qy
            double_plane['Y'].append((j,k,l,m))  # e.g. f0111, f0120
        elif l == 0: # same, but for the horizontal plane
            double_plane['X'].append((j,k,l,m))  # e.g. f1001, f1002
        else:  # The RDT can be seen on both planes and is a multiple of both Qx and Qy
            double_plane['X'].append((j,k,l,m))  # e.g. f1020, f1120
            double_plane['Y'].append((j,k,l,m))

    return single_plane, double_plane


def calculate(measure_input, input_files, tunes, phases, invariants, header):
    """

    Args:
        measure_input:
        input_files:
        tunes:
        invariants:
        header:

    Returns:
    """
    LOGGER.info(f"Start of RDT analysis")
    meas_input = deepcopy(measure_input)
    meas_input["compensation"] = "none"
    LOGGER.info(f"Calculating RDTs up to magnet order {meas_input['rdt_magnet_order']}")

    single_plane_rdts, double_plane_rdts = _generate_plane_rdts(meas_input["rdt_magnet_order"])
    for plane in PLANES:
        bpm_names = input_files.bpms(plane=plane, dpp_value=0)
        for_rdts = _best_90_degree_phases(meas_input, bpm_names, phases, tunes, plane)
        LOGGER.info(f"Average phase advance between BPM pairs: {for_rdts.loc[:,'MEAS'].mean()}")
        for rdt in single_plane_rdts[plane]:
            try:
                df = _process_rdt(meas_input, input_files, for_rdts, invariants, plane, rdt)
            except ValueError as e:  # catch the AMP line not being found in the lin file
                LOGGER.warning(f"RDT calculation failed for {jklm2str(*rdt)}: {str(e)}")
                continue
            write(df, add_freq_to_header(header, plane, rdt), meas_input, plane, rdt)
    for plane in PLANES:
        bpm_names = input_files.bpms(dpp_value=0)
        for_rdts = _best_90_degree_phases(meas_input, bpm_names, phases, tunes, plane)
        LOGGER.info(f"Average phase advance between BPM pairs: {for_rdts.loc[:, 'MEAS'].mean()}")
        for rdt in double_plane_rdts[plane]:
            try:
                df = _process_rdt(meas_input, input_files, for_rdts, invariants, plane, rdt)
            except ValueError as e:
                LOGGER.warning(f"RDT calculation failed for {jklm2str(*rdt)}: {str(e)}")
                continue
            write(df, add_freq_to_header(header, plane, rdt), meas_input, plane, rdt)


def write(df, header, meas_input, plane, rdt):
    outputdir = join(meas_input.outputdir, "rdt", _rdt_to_order_and_type(rdt))
    iotools.create_dirs(outputdir)
    tfs.write(join(outputdir, f"f{_rdt_to_str(rdt)}_{plane.lower()}{EXT}"), df, header,
              save_index="NAME")


def _rdt_to_str(rdt):
    j, k, l, m = rdt
    return f"{j}{k}{l}{m}"


def _rdt_to_order_and_type(rdt):
    j, k, l, m = rdt
    rdt_type = "normal" if (l + m) % 2 == 0 else "skew"
    orders = dict(((1, "dipole"), 
                   (2, "quadrupole"), 
                   (3, "sextupole"), 
                   (4, "octupole"),
                   (5, "decapole"),
                   (6, "dodecapole"),
                   (7, "tetradecapole"),
                   (8, "hexadecapole"),
                 ))
    return f"{rdt_type}_{orders[j + k + l + m]}"


def _best_90_degree_phases(meas_input, bpm_names, phases, tunes, plane):
    filtered = phases[plane]["uncompensated"]["MEAS"].loc[bpm_names, bpm_names]
    phase_meas = pd.concat(
        (filtered % 1, (filtered.iloc[:, :NBPMS_FOR_90] + tunes[plane]["Q"]) % 1), axis=1)
    second_bmps = np.abs(phase_meas * _get_n_upper_diagonals(NBPMS_FOR_90, phase_meas.shape)
                         - 0.25).idxmin(axis=1)
    filtered.iloc[-NBPMS_FOR_90:, :NBPMS_FOR_90] = (filtered.iloc[-NBPMS_FOR_90:,
                                                    :NBPMS_FOR_90] + tunes[plane]["Q"]) % 1
    filtered["NAME2"], filtered["MEAS"], filtered["ERRMEAS"] = second_bmps, filtered.lookup(
        bpm_names, second_bmps), phases[plane]["uncompensated"]["ERRMEAS"].lookup(bpm_names, second_bmps)
    for_rdts = pd.merge(filtered.loc[:, ["NAME2", "MEAS", "ERRMEAS"]],
                        meas_input.accelerator.model.loc[:, ["S"]], how="inner",
                        left_index=True, right_index=True)
    return for_rdts


def _get_n_upper_diagonals(n, shape):
    return diags(np.ones((n, shape[0])), np.arange(n)+1, shape=shape).toarray()


def _determine_line(rdt, plane):
    j, k, l, m = rdt
    lines = dict(X=(1 - j + k, m - l, 0),
                 Y=(k - j, 1 - l + m, 0))
    return lines[plane]


def add_freq_to_header(header, plane, rdt):
    mod_header = header.copy()
    line = _determine_line(rdt, plane)
    freq = np.mod(line@np.array([header['Q1'], header['Q2'], 0]), 1)
    mod_header["FREQ"] = freq if freq <= 0.5 else 1 - freq
    return mod_header


def _process_rdt(meas_input, input_files, phase_data, invariants, plane, rdt):
    df = pd.DataFrame(phase_data)
    second_bpms = df.loc[:, "NAME2"].to_numpy()
    df["S2"] = df.loc[second_bpms, "S"].to_numpy()
    df["COUNT"] = len(input_files.dpp_frames(plane, 0))
    line = _determine_line(rdt, plane)
    phase_sign, suffix = get_line_sign_and_suffix(line, input_files, plane)
    comp_coeffs1 = to_complex(
        input_files.joined_frame(plane, [f"AMP{suffix}"], dpp_value=0).loc[df.index, :].to_numpy(),
        phase_sign * input_files.joined_frame(plane, [f"PHASE{suffix}"], dpp_value=0).loc[df.index, :].to_numpy())
    # Multiples of tunes needs to be added to phase at second BPM if that is in second turn
    phase2 = phase_sign * input_files.joined_frame(plane, [f"PHASE{suffix}"], dpp_value=0).loc[second_bpms, :].to_numpy()
    comp_coeffs2 = to_complex(
        input_files.joined_frame(plane, [f"AMP{suffix}"], dpp_value=0).loc[second_bpms, :].to_numpy(),
        _add_tunes_if_in_second_turn(df, input_files, line, phase2))
    # Get amplitude and phase of the line from linx/liny file
    line_amp, line_phase, line_amp_e, line_phase_e = complex_secondary_lines(  # TODO use the errors
        df.loc[:, "MEAS"].to_numpy()[:, np.newaxis] * meas_input.accelerator.beam_direction,
        df.loc[:, "ERRMEAS"].to_numpy()[:, np.newaxis], comp_coeffs1, comp_coeffs2)
    rdt_phases_per_file = _calculate_rdt_phases_from_line_phases(df, input_files, line, line_phase)
    rdt_angles = stats.circular_mean(rdt_phases_per_file, period=1, axis=1) % 1
    df[f"PHASE"] = rdt_angles
    df[f"{ERR}PHASE"] = stats.circular_error(rdt_phases_per_file, period=1, axis=1)
    df[AMPLITUDE], df[f"{ERR}{AMPLITUDE}"] = _fit_rdt_amplitudes(invariants, line_amp, plane, rdt)
    df[f"REAL"] = np.cos(2 * np.pi * rdt_angles) * df.loc[:, AMPLITUDE].to_numpy()
    df[f"IMAG"] = np.sin(2 * np.pi * rdt_angles) * df.loc[:, AMPLITUDE].to_numpy()
    # in old files there was "EAMP" and "PHASE_STD"
    return df.loc[:, ["S", "COUNT", AMPLITUDE, f"{ERR}{AMPLITUDE}", "PHASE", f"{ERR}PHASE", "REAL", "IMAG"]]


def _add_tunes_if_in_second_turn(df, input_files, line, phase2):
    mask = df_diff(df, "S", "S2") > 0
    tunes = np.empty((2, len(input_files.dpp_frames("X", 0))))
    for i, plane in enumerate(PLANES):
        tunes[i] = np.array([lin.headers[f"Q{i+1}"] for lin in input_files.dpp_frames(plane, 0)])
    phase2[mask, :] = phase2[mask, :] + line[0] * tunes[0] + line[1] * tunes[1]
    return phase2


def _calculate_rdt_phases_from_line_phases(df, input_files, line, line_phase):
    phases = np.zeros((2, df.index.size, len(input_files.dpp_frames("X", 0))))
    for i, plane in enumerate(PLANES):
        if line[i] != 0:
            phases[i] = input_files.joined_frame(plane, [f"MU{plane}"], dpp_value=0).loc[df.index, :].to_numpy() % 1
    return line_phase - line[0] * phases[0] - line[1] * phases[1] + 0.25


def _fit_rdt_amplitudes(invariants, line_amp, plane, rdt):
    """
    Returns RDT amplitudes in units of meters ^ {1 - n/2}, where n is the order of RDT.
    """
    amps, err_amps = np.empty(line_amp.shape[0]), np.empty(line_amp.shape[0])
    kick_data = get_linearized_problem(invariants, plane, rdt)  # corresponding to actions in meters
    guess = np.mean(line_amp / kick_data, axis=1)

    def fitting(x, f):
        return f * x

    for i, bpm_rdt_data in enumerate(line_amp):
        popt, pcov = curve_fit(fitting, kick_data, bpm_rdt_data, p0=guess[i])
        amps[i] = popt[0]
        err_amps[i] = np.sqrt(pcov)[0] if np.isfinite(np.sqrt(pcov)[0]) else 0. # if single file is used, the error is reported as Inf, which is then overwritten with 0
    return amps, err_amps


def get_linearized_problem(invs, plane, rdt):
    """
    2 * j * f_jklm * (powers of 2Jx and 2Jy) : f_jklm is later a parameter of a fit
    we use sqrt(2J): unit is sqrt(m).
    """
    j, k, l, m = rdt
    act_x = invs["X"].T[0]
    act_y = invs["Y"].T[0]
    if plane == "X":
        return 2 * j * act_x ** (j + k - 2) * act_y ** (l + m)
    return 2 * l * act_x ** (j + k) * act_y ** (l + m - 2)


def get_line_sign_and_suffix(line, input_files, plane):
    suffix = f"{line[0]}{line[1]}".replace("-", "_")
    conj_suffix = f"{-line[0]}{-line[1]}".replace("-", "_")
    if f"AMP{suffix}" in input_files[plane][0].columns:
        return 1, suffix
    if f"AMP{conj_suffix}" in input_files[plane][0].columns:
        return -1, conj_suffix

    # The specified AMP column hasn't been found in the lin file
    msg = (f"The column AMP{suffix} has not been found in the lin{plane.lower()} file. "
            "Consider re-running the frequency analysis with a higher order resonance term")
    raise ValueError(msg)


def complex_secondary_lines(phase_adv, err_padv, sig1, sig2):
    """

    Args:
        phase_adv: phase advances between two BPMs.
        err_padv: error on the phase advance between two BPMs.
        sig1: Complex coefficients of a secondary lines at the first BPM of the pairs.
        sig2: Complex coefficients of a secondary lines at the second BPM of the pairs.

    Returns:
         `Tuple` with amplitudes, phases err_amplitudes and err_phases of the complex signal.
    """
    tp = 2.0 * np.pi
    # computing complex secondary line (h-)
    sig = (sig1 * (1 + 1j / np.tan(phase_adv * tp)) - sig2 * 1j / np.sin(phase_adv * tp)) / 2
    # computing error secondary line (h-) # TODO is this really the error?
    esig = (sig1 * 1j / np.square(np.sin(phase_adv * tp)) +
            sig2 * -1j * np.cos(phase_adv * tp) / np.square(np.sin(phase_adv * tp))) * err_padv / 2
    return (np.abs(sig), (np.angle(sig) / tp) % 1,
            np.abs(esig), (np.angle(esig) / tp) % 1)


def to_complex(amplitudes, phases, period=1):
    return amplitudes * np.exp(2j * np.pi * phases / period)



"""
Plot TFS
--------

Easily plot tfs-files with all kinds of additional functionality and ways to combine plots.

.. code-block:: python

    from omc3.plotting.plot_tfs import plot

    figs = plot(
        files=['beta_phase_{0}.tfs', 'beta_amp_{0}.tfs'],
        same_figure='planes',
        same_axes=['files'],
        x_columns=['S'],
        y_columns=['BET{0}'],
        error_columns=None,
        planes=['X', 'Y'],
        x_labels=['Location [m]'],
        file_labels=[r'$\\beta$ from phase', r'$\beta$ from amplitude'],
        # column_labels=[r'$\\beta_{0}$'],    # would have bx in legend
        column_labels=[''],                  # removes COLUMNX COLUMNY from legend-names
        y_labels=[[r'$\\beta_x$', r'$\\beta_y$']],  # axes labels (outer = figures, inner = axes)
        output='output_dir',
        show=False,
        single_legend=True,
        change_marker=True,
    )


**Arguments:**

*--Required--*

- **files** *(MultiClass)*: Path to files to plot.
  If planes are used, replace the plane in the filename with '{0}'

- **x_columns** *(str)*: List of column names to use as x-values.

- **y_columns** *(str)*: List of column names to plot (e.g. BETX, BETY or BET{0} if `planes` is used.)


*--Optional--*

- **change_marker**: Changes marker for each line in the plot.

  Action: ``store_true``
- **column_labels** *(str)*: Column-Labels for the plots, default: y_columns.

- **error_columns** *(str)*: List of parameters to get error values from.

- **errorbar_alpha** *(float)*: Alpha value for error bars

  Default: ``0.6``
- **file_labels** *(str)*: Labels for the files, default: filenames.

- **manual_style** *(DictAsString)*: Additional style rcParameters which update the set of predefined ones.

  Default: ``{}``
- **ncol_legend** *(int)*: Number of bpm legend-columns. If < 1 no legend is shown.

  Default: ``3``
- **output** *(MultiClass)*: Folder to output the plots to.

- **output_prefix** *(str)*: Prefix for the output filename.

  Default: ``plot_``
- **planes** *(str)*: Works only with filenames ending in 'x' and 'y' and columns ending in X or Y.
  These suffixes will be attached to the given files and y_columns.

  Choices: ``('X', 'Y')``
- **plot_styles** *(str)*: Which plotting styles to use,
  either from plotting.styles.*.mplstyles or default mpl.

  Default: ``['standard']``
- **same_axes** *(str)*: Combine plots into single axes. Multiple choices possible.

  Choices: ``['files', 'columns', 'planes']``
- **same_figure** *(str)*: Plot two axes into the same figure (can't be the same as 'same_axes').
  Has no effect if there is only one of the given thing.

  Choices: ``['files', 'columns', 'planes']``
- **share_xaxis**: In case of multiple axes per figure, share x-axis.

  Action: ``store_true``
- **show**: Shows plots.

  Action: ``store_true``
- **single_legend**: Show only one legend instance (at the top plot).

  Action: ``store_true``
- **vertical_lines** *(DictAsString)*: List of vertical lines (e.g. IR positions) to plot.
  Need to contain arguments for axvline, and may contain the additional keys "text" and "loc"
  which is one of  ['bottom', 'top', 'line bottom', 'line top'] and places the text at the given location.

  Default: ``[]``
- **x_labels** *(str)*: Labels for the x-axis, default: x_columns.

- **x_lim** *(float, int, None)*: Limits on the x axis (Tupel)

- **y_labels**: Override labels for the y-axis, default: file_labels or column_labels (depending on same_axes).
  Needs to be a list of lists, where the inner list goes over the axes in one figure and the outer over the figures.
  If the respective length is 1, the same label will be used for all figures or axes.

- **y_lim** *(float, int None)*: Limits on the y axis (Tupel)
"""
from collections import OrderedDict
from pathlib import Path

import matplotlib
import tfs
from generic_parser import EntryPointParameters, entrypoint, DotDict
from generic_parser.entry_datatypes import DictAsString, get_multi_class
from matplotlib import pyplot as plt, rcParams

from omc3.definitions.constants import PLANES
from omc3.optics_measurements.constants import EXT
from omc3.plotting.optics_measurements.constants import DEFAULTS
from omc3.plotting.optics_measurements.utils import FigureCollector
from omc3.plotting.spectrum.utils import get_unique_filenames, output_plot
from omc3.plotting.utils import (annotations as pannot, lines as plines,
                                 style as pstyle, colors as pcolors)
from omc3.plotting.utils.lines import VERTICAL_LINES_TEXT_LOCATIONS
from omc3.utils.iotools import PathOrStr, save_config
from omc3.utils.logging_tools import get_logger, list2str

LOG = get_logger(__name__)

float_or_none = get_multi_class(float, int, type(None))  # for the limits


def get_params():
    params = EntryPointParameters()
    params.add_parameter(
        name="files",
        help=("Path to files to plot. "
              "If planes are used, replace the plane in the filename with '{0}'"),
        required=True,
        nargs="+",
        type=PathOrStr,
    )
    params.add_parameter(
        name="y_columns",
        help="List of column names to plot (e.g. BETX, BETY or BET{0} if `planes` is used.)",
        required=True,
        type=str,
        nargs="+",
    )
    params.add_parameter(
        name="x_columns",
        help="List of column names to use as x-values.",
        required=True,
        type=str,
        nargs="+",
    )
    params.add_parameter(
        name="error_columns",
        help="List of parameters to get error values from.",
        type=str,
        nargs="+",
    )
    params.add_parameter(
        name="column_labels",
        help="Column-Labels for the plots, default: y_columns.",
        type=str,
        nargs="+",
    )
    params.add_parameter(
        name="file_labels",
        help="Labels for the files, default: filenames.",
        type=str,
        nargs="+",
    )
    params.add_parameter(
        name="x_labels",
        help="Labels for the x-axis, default: x_columns.",
        type=str,
        nargs="+",
    )
    params.add_parameter(
        name="y_labels",
        help=("Override labels for the y-axis, default: file_labels or column_labels "
              "(depending on same_axes). Needs to be a list of lists, where the "
              "inner list goes over the axes in one figure and the outer over "
              "the figures. If the respective length is 1, the same label will "
              "be used for all figures or axes."
              ),
        nargs='+',
    )
    params.add_parameter(
        name="planes",
        help=("Works only with filenames ending in 'x' and 'y' and "
              "columns ending in X or Y. These suffixes will be attached "
              "to the given files and y_columns."),
        type=str,
        nargs='+',
        choices=PLANES,
    )
    params.add_parameter(
        name="output",
        help="Folder to output the plots to.",
        type=PathOrStr,
    )
    params.add_parameter(
        name="output_prefix",
        help="Prefix for the output filename.",
        type=str,
        default="plot_"
    )
    params.add_parameter(
        name="show",
        help="Shows plots.",
        action="store_true",
    )
    params.add_parameter(
        name="same_axes",
        help="Combine plots into single axes. Multiple choices possible.",
        type=str,
        nargs='+',
        choices=['files', 'columns', 'planes']
    )
    params.add_parameter(
        name="same_figure",
        help=("Plot two axes into the same figure "
              "(can't be the same as 'same_axes'). "
              "Has no effect if there is only one of the given thing."),
        type=str,
        choices=['files', 'columns', 'planes']
    )
    params.add_parameter(
        name="single_legend",
        help="Show only one legend instance (at the top plot).",
        action="store_true",
    )
    params.add_parameter(
        name="x_lim",
        nargs=2,
        type=float_or_none,
        help='Limits on the x axis (Tupel)'
    )
    params.add_parameter(
        name="y_lim",
        nargs=2,
        type=float_or_none,
        help='Limits on the y axis (Tupel)'
    )
    params.add_parameter(
        name="vertical_lines",
        nargs="*",
        default=[],
        type=DictAsString,
        help='List of vertical lines (e.g. IR positions) to plot. '
             'Need to contain arguments for axvline, and may contain '
             'the additional keys "text" and "loc" which is one of '
             f' {list(VERTICAL_LINES_TEXT_LOCATIONS.keys())} and places the text at the given location.')

    # Plotting Style Parameters ---
    params.add_parameter(
        name="plot_styles",
        type=str,
        nargs="+",
        default=['standard'],
        help='Which plotting styles to use, either from plotting.styles.*.mplstyles or default mpl.'
    )
    params.add_parameter(
        name="manual_style",
        type=DictAsString,
        default={},
        help='Additional style rcParameters which update the set of predefined ones.'
    )
    params.add_parameter(
        name="change_marker",
        help="Changes marker for each line in the plot.",
        action="store_true",
    )
    params.add_parameter(
        name="ncol_legend",
        type=int,
        default=DEFAULTS['ncol_legend'],
        help='Number of bpm legend-columns. If < 1 no legend is shown.'
    )
    params.add_parameter(
        name="errorbar_alpha",
        help="Alpha value for error bars",
        type=float,
        default=DEFAULTS['errorbar_alpha'],
    )
    params.add_parameter(
        name="share_xaxis",
        help="In case of multiple axes per figure, share x-axis.",
        action="store_true",
    )
    return params


# Main invocation ############################################################


@entrypoint(get_params(), strict=True)
def plot(opt):
    """Main plotting function."""
    LOG.info(f"Starting plotting of tfs files: {list2str(opt.files):s}")
    if opt.output is not None:
        save_config(Path(opt.output), opt, __file__)

    # preparations
    opt = _check_opt(opt)
    pstyle.set_style(opt.plot_styles, opt.manual_style)

    # extract data
    fig_collection = sort_data(opt.get_subdict(
        ['files', 'planes', 'x_columns', 'y_columns', 'error_columns',
         'file_labels', 'column_labels', 'x_labels', 'y_labels',
         'same_axes', 'same_figure', 'output', 'output_prefix',
         'share_xaxis']))

    # plotting
    _create_plots(fig_collection,
                  opt.get_subdict(['x_lim', 'y_lim',
                                   'ncol_legend', 'single_legend',
                                   'change_marker', 'errorbar_alpha',
                                   'vertical_lines', 'show']))

    return fig_collection.fig_dict


# Sorting ----------------------------------------------------------------------


def sort_data(opt):
    """Load all data from files and sort into figures."""
    collector = FigureCollector()
    axes_ids = _get_axes_ids(opt)

    if len(axes_ids) == 1 and opt.share_xaxis:
        LOG.warning("Option `share_xaxis` is activated but only one axes is present. The option has no effect.")

    same_axes_set = frozenset()
    if opt.same_axes:
        same_axes_set = frozenset(opt.same_axes)

    for (file_path, filename), file_label in zip(get_unique_filenames(opt.files), opt.file_labels):
        for x_col, y_col, err_col, x_label, column_label in zip(
                opt.x_columns, opt.y_columns, opt.error_columns, opt.x_labels, opt.column_labels):

            if opt.planes is None:
                id_map = get_id(filename, y_col, file_label, column_label, same_axes_set,
                                opt.same_figure, opt.output_prefix)
                output_path = _get_full_output_path(opt.output, id_map['figure_id'])

                collector.add_data_for_id(
                    figure_id=id_map['figure_id'],
                    label=id_map['legend_label'],
                    data=_read_data(file_path, x_col, y_col, err_col),
                    path=output_path,
                    x_label=x_label,
                    y_label=id_map['ylabel'],
                    axes_id=id_map['axes_id'],
                    axes_ids=axes_ids,
                )

            else:
                for plane in opt.planes:
                    id_map = get_id(filename, y_col, file_label, column_label, same_axes_set,
                                    opt.same_figure, opt.output_prefix, plane, opt.planes)
                    output_path = _get_full_output_path(opt.output, id_map['figure_id'])

                    file_path_plane = file_path.with_name(f'{_safe_format(file_path.name, plane.lower())}')
                    y_col_plane = _safe_format(y_col, plane.upper())
                    err_col_plane = _safe_format(err_col, plane.upper())
                    x_col_plane = _safe_format(x_col, plane.upper())

                    collector.add_data_for_id(
                        figure_id=id_map['figure_id'],
                        label=id_map['legend_label'],
                        data=_read_data(file_path_plane, x_col_plane, y_col_plane, err_col_plane),
                        path=output_path,
                        x_label=x_label,
                        y_label=id_map['ylabel'],
                        axes_id=id_map['axes_id'],
                        axes_ids=axes_ids,
                    )
    if opt.y_labels:
        collector = _update_y_labels(collector, opt.y_labels)

    if opt.share_xaxis:
        collector = _share_xaxis(collector)

    return collector


def get_id(filename_parts, column, file_label, column_label, same_axes, same_figure, prefix, plane='', planes=[]):
    """
    Get the right IDs for the current sorting way.
    This is where the actual sorting happens, by mapping the right IDs according to the chosen
    options.
    """
    planes = "".join(planes)

    file_last = filename_parts[-1].replace(EXT, "").strip("_")
    file_output = "_".join(filename_parts).replace(EXT, "").strip("_")
    if file_label is not None:
        file_output = f'{file_label}_{file_last}'
    else:
        file_label = file_output
    file_output_planes = _safe_format(file_output, planes.lower())
    file_output = _safe_format(file_output, plane)

    column_planes = _safe_format(column, planes)
    column = _safe_format(column, plane)
    if column_label is None:
        column_label = column
    else:
        column_label = _safe_format(column_label, plane.lower())

    axes_id = {'files': '_'.join(filename_parts),
               'columns': f'{column}',
               'planes': f'{plane.lower()}'
               }.get(same_figure, '')

    key = same_axes if same_figure is None else same_axes.union({same_figure})

    out_dict = {
        frozenset(['files', 'columns', 'planes']): dict(
            figure_id=f'{prefix}{planes}',
            axes_id=axes_id,
            legend_label=f'{file_label} {column_label}',
            ylabel=f'',
        ),
        frozenset(['files', 'columns']): dict(
            figure_id=f'{prefix}{plane.lower()}',
            axes_id=axes_id,
            legend_label=f'{file_label} {column_label}',
            ylabel=f'',
        ),
        frozenset(['planes', 'columns']): dict(
            figure_id=f'{prefix}{file_output_planes}',
            axes_id=axes_id,
            legend_label=column_label,
            ylabel=column_label,
        ),
        frozenset(['planes', 'files']): dict(
            figure_id=f'{prefix}{column_planes}',
            axes_id=axes_id,
            legend_label=f'{file_label} {column_label}',
            ylabel=column_label,
        ),
        frozenset(['planes']): dict(
            figure_id=f'{prefix}{file_output_planes}_{column_planes}',
            axes_id=axes_id,
            legend_label=column_label,
            ylabel=column_label,
        ),
        frozenset(['files']): dict(
            figure_id=f'{prefix}{column}',
            axes_id=axes_id,
            legend_label=f'{file_label}',
            ylabel=column_label,
        ),
        frozenset(['columns']): dict(
            figure_id=f'{prefix}{file_output}',
            axes_id=axes_id,
            legend_label=column_label,
            ylabel=f'{file_label}',
        ),
        frozenset([]): dict(
            figure_id=f'{prefix}{file_output}_{column}',
            axes_id=axes_id,
            legend_label=f'',
            ylabel=column_label,
        ),
    }[key]

    out_dict['figure_id'] = out_dict['figure_id'].strip("_")
    return out_dict


def _read_data(file_path, x_col, y_col, err_col):
    tfs_data = tfs.read(str(file_path))
    return DotDict(
        x=tfs_data[x_col],
        y=tfs_data[y_col],
        err=tfs_data[err_col] if err_col is not None else None,
    )


def _update_y_labels(fig_collection, y_labels):
    if len(y_labels) == 1:
        y_labels = y_labels * len(fig_collection.figs)

    for idx_fig, fig_container in enumerate(fig_collection.figs.values()):
        axes_labels = y_labels[idx_fig]
        if len(axes_labels) == 1:
            axes_labels = axes_labels * len(fig_container.ylabels)

        for idx_ax, (ax_id, _) in enumerate(fig_container.ylabels.items()):
            fig_container.ylabels[ax_id] = _safe_format(axes_labels[idx_ax], ax_id)
    return fig_collection


def _share_xaxis(fig_collection):
    """Shared xaxis at last axes and remove all xlabels, ticks of other axes."""
    for fig_container in fig_collection.figs.values():
        axs = fig_container.axes.values()
        fig_container.axes[fig_container.axes_ids[-1]].get_shared_x_axes().join(*axs)
        for ax_id in fig_container.axes_ids[:-1]:
            fig_container.axes[ax_id].set_xticklabels([])
            fig_container.xlabels[ax_id] = None

    return fig_collection


# Plotting ---------------------------------------------------------------------


def _create_plots(fig_collection, opt):
    """Main plotting routine."""
    for fig_container in fig_collection.figs.values():
        for idx_ax, ax_id in enumerate(fig_container.axes_ids):
            ax, data, xlabel, ylabel = fig_container[ax_id]
            _plot_vlines(ax, opt.vertical_lines)
            _plot_data(ax, data, opt.change_marker, opt.errorbar_alpha)
            _set_axes_layout(ax, opt.x_lim, opt.y_lim, ylabel, xlabel)

            # plt.show(block=False)  # for debugging
            if idx_ax == 0 or not opt.single_legend:
                pannot.make_top_legend(ax, opt.ncol_legend)

        output_plot(fig_container)

    if opt.show:
        plt.show()


def _plot_data(ax, data, change_marker, ebar_alpha):
    for idx, (label, values) in enumerate(data.items()):
        ebar = ax.errorbar(values.x, values.y, yerr=values.err,
                           ls=rcParams[u"lines.linestyle"],
                           fmt=_get_marker(idx, change_marker),
                           label=label)

        pcolors.change_ebar_alpha_for_line(ebar, ebar_alpha)


def _plot_vlines(ax, lines):
    if not lines:
        return

    for line in lines:
        loc = line.pop('loc', None)
        text = line.pop('text', None)
        plines.plot_vertical_line(ax, line, text=text, text_loc=loc)
        line['loc'] = loc  # put back for other axes
        line['text'] = text


def _set_axes_layout(ax, x_lim, y_lim, ylabel, xlabel):
    ax.set_xlim(x_lim)
    ax.set_xlabel(xlabel)
    ax.set_ylim(y_lim)
    ax.set_ylabel(ylabel)


# Output ---


def _get_full_output_path(folder, filename):
    if folder is None or filename is None:
        return None
    return Path(folder) / f"{filename}.{matplotlib.rcParams['savefig.format']}"


# Helper -----------------------------------------------------------------------


def _check_opt(opt):
    """Sanity checks for the opt structure and broadcasting of parameters."""
    if opt.file_labels is None:
        opt.file_labels = [None] * len(opt.files)
    elif len(opt.file_labels) != len(opt.files):
        raise AttributeError("The number of file-labels and number of files differ!")

    if len(opt.x_columns) == 1 and len(opt.y_columns) > 1:
        opt.x_columns *= len(opt.y_columns)
    elif len(opt.x_columns) != len(opt.y_columns):
        raise AttributeError("The number of x-columns and y-columns differ!")

    if opt.x_labels is None:
        opt.x_labels = [None] * len(opt.x_columns)
    elif len(opt.x_labels) == 1 and len(opt.x_columns) > 1:
        opt.x_labels *= len(opt.x_columns)
    elif len(opt.x_labels) != len(opt.x_columns):
        raise AttributeError("The number of x-labels and x-columns differ!")

    if opt.error_columns is None:
        opt.error_columns = [None] * len(opt.y_columns)
    elif len(opt.error_columns) != len(opt.y_columns):
        raise AttributeError("The number of error-columns and number of y-columns differ!")

    if opt.column_labels is None:
        opt.column_labels = [None] * len(opt.y_columns)
    elif len(opt.column_labels) == 1 and len(opt.y_columns) > 1:
        opt.column_labels *= len(opt.y_columns)
    elif len(opt.column_labels) != len(opt.y_columns):
        raise AttributeError("The number of column-labels and number of y columns differ!")

    if opt.same_figure is not None:
        if opt.same_axes is not None and opt.same_figure in opt.same_axes:
            raise AttributeError("Found the same option in 'same_axes' "
                                 "and 'same_figure'. This is not allowed.")

    return opt


def _get_axes_ids(opt):
    """
    Get's all id's first (to know how many axes to use). So the `get_id` is done later for
    actual plotting again.
    The order matters, as this function determines the order in which the data is distributed on
    the figure axes (if multiple), which is important if one gives the manual y-labels option.
    Couldn't find a quicker way... (jdilly)
    """
    axes_ids = []  # needs to be something ordered
    same_axes_set = frozenset()
    if opt.same_axes:
        same_axes_set = frozenset(opt.same_axes)
    for (file_path, filename), file_label in zip(get_unique_filenames(opt.files), opt.file_labels):
        for x_col, y_col, err_col, x_label, column_label in zip(
                opt.x_columns, opt.y_columns, opt.error_columns, opt.x_labels, opt.column_labels):

            if opt.planes is None:
                id_ = get_id(filename, y_col, file_label, column_label, same_axes_set,
                             opt.same_figure, opt.output_prefix)['axes_id']
                axes_ids.append(id_)
            else:
                for plane in opt.planes:
                    id_ = get_id(filename, y_col, file_label, column_label, same_axes_set,
                                 opt.same_figure, opt.output_prefix, plane, opt.planes)['axes_id']
                    axes_ids.append(id_)
    return list(OrderedDict.fromkeys(axes_ids).keys())  # unique list with preserved order


def _get_marker(idx, change):
    """Return the marker used"""
    if change:
        return plines.MarkerList.get_marker(idx)
    else:
        return rcParams['lines.marker']


def _safe_format(label, insert):
    """
    Formats label. Usually just ``.format()`` works fine, even if there is no {}-placeholder in
    the label, but it might cause errors if latex is used in the label or ``None``. Hence the
    try-excepts.
    """
    try:
        return label.format(insert)
    except KeyError:  # can happen for latex strings
        return label
    except AttributeError:  # label is None
        return None


# Script Mode ------------------------------------------------------------------


if __name__ == "__main__":
    plot()



"""
Filters
-------

Filters for different kind of measurement data, and to filter the entries
in the response matrix based on (the presumably then filtered) measurement data.
Measurement filters extract valid (or trustworthy) data, e.g. to be used in
corrections.
The main function is :meth:`omc3.correction.filters.filter_measurement`
which decides on which filter to use for the given keys.

In earlier implementations there was a split between all kinds of measures,
i.e. beta, phase etc. In this implementation most of it is handled by
the `_get_filtered_generic` function.
"""
from collections import defaultdict
from typing import Callable, Dict, Sequence

import numpy as np
import pandas as pd
import tfs
from generic_parser import DotDict

from omc3.correction.constants import DELTA, ERR, ERROR, NAME2, PHASE, PHASE_ADV, TUNE, VALUE, WEIGHT
from omc3.definitions.constants import PLANES
from omc3.optics_measurements.constants import AMPLITUDE, F1001, F1010, IMAG, REAL
from omc3.utils import logging_tools, stats

LOG = logging_tools.get_logger(__name__)


# Measurement Filter -----------------------------------------------------------


def filter_measurement(
    keys: Sequence[str], meas: Dict[str, pd.DataFrame], model: pd.DataFrame, opt: DotDict
) -> dict:
    """Filters measurements in `keys` based on the dict-entries (keys as in `keys`)
    in `opt.errorcut`, `opt.modelcut` and `opt.weights` and unifies the
    data-column names to VALUE, ERROR, WEIGHT.
    If `opt.use_errorbars` is `True` the weights will be also based on the errors."""
    filters = _get_measurement_filters()
    new = dict.fromkeys(keys)
    for key in keys:
        new[key] = filters[key](key, meas[key], model, opt)
    return new


def _get_measurement_filters() -> defaultdict:
    """Returns a dict with the respective `_get_*` filter-functions that defaults
    to `~_get_filtered_generic`. Some columns might need to have extra steps for
    filtering, or simply a different filtering process."""
    return defaultdict(
        lambda: _get_filtered_generic,
        {
            f"{TUNE}": _get_tunes,
            f"{F1010}I": _get_coupling,
            f"{F1010}R": _get_coupling,
            f"{F1001}I": _get_coupling,
            f"{F1001}R": _get_coupling,
        },
    )


def _get_filtered_generic(col: str, meas: pd.DataFrame, model: pd.DataFrame, opt: DotDict) -> tfs.TfsDataFrame:
    """
    Filters the provided column *col* of the measurement dataframe *meas*, based on the model values
    (from the *model* dataframe) and the filtering options given at the command line (for instance,
    the ``errorcut`` and ``modelcut`` values).

    Args:
        col (str): The column name to filter.
        meas (pd.DataFrame): The measurement dataframe
        model (pd.DataFrame): The model dataframe, which we get from the ``model_creator``.
        opt (DotDict): The command line options dictionary.

    Returns:
        The filtered dataframe as a `~tfs.TfsDataFrame`.
    """
    common_bpms = meas.index.intersection(model.index)
    meas = meas.loc[common_bpms, :]

    new = tfs.TfsDataFrame(index=common_bpms)
    new[VALUE] = meas.loc[:, col].to_numpy()
    new[ERROR] = meas.loc[:, f"{ERR}{col}"].to_numpy()
    new[WEIGHT] = (
        _get_errorbased_weights(col, opt.weights[col], meas.loc[:, f"{ERR}{DELTA}{col}"])
        if opt.use_errorbars
        else opt.weights[col]
    )

    # Applying filtering cuts
    error_filter = meas.loc[:, f"{ERR}{DELTA}{col}"].to_numpy() < opt.errorcut[col]
    model_filter = np.abs(meas.loc[:, f"{DELTA}{col}"].to_numpy()) < opt.modelcut[col]
    # if opt.automatic_model_cut:  # TODO automated model cut
    #     model_filter = _get_smallest_data_mask(np.abs(meas.loc[:, f"{DELTA}{col}"].to_numpy()), portion=0.95)
    if f"{PHASE}" in col:
        new[NAME2] = meas.loc[:, NAME2].to_numpy()
        second_bpm_in = np.in1d(new.loc[:, NAME2].to_numpy(), new.index.to_numpy())
        good_bpms = error_filter & model_filter & second_bpm_in
        good_bpms[-1] = False
    else:
        good_bpms = error_filter & model_filter
    LOG.debug(f"Number of BPMs kept for column '{col}' after filtering: {np.sum(good_bpms)}")
    return new.loc[good_bpms, :]


def _get_tunes(key: str, meas: pd.DataFrame, model, opt: DotDict):
    meas[WEIGHT] = opt.weights[key]
    if opt.use_errorbars:
        meas[WEIGHT] = _get_errorbased_weights(key, meas[WEIGHT], meas[ERROR])
    LOG.debug(f"Number of tune measurements: {len(meas.index.to_numpy())}")
    return meas


def _get_coupling(col: str, meas: pd.DataFrame, model: pd.DataFrame, opt: DotDict) -> tfs.TfsDataFrame:
    """
    Applies filters to the coupling dataframe *meas*. This is a bit hacky. Takes the measurement and
    model dataframes for one of the coupling RDTs (*meas* comes from **f1001.tfs** or **f1010.tfs**)
    maps the column *col* name to the "old" naming (for the F1001 for instance, REAL -> F1001R)
    before passing to `~_get_filtered_generic` which will do the filtering. This is because
    `~_get_filtered_generic` does a comparison to the model which has old names.

    Args:
        col (str): The column name to filter.
        meas (pd.DataFrame): The measurement dataframe, which here will be the loaded **f1001.tfs**
            or **f1010.tfs** file.
        model (pd.DataFrame): The model dataframe, which we get from the ``model_creator``.
        opt (DotDict): The command line options dictionary.

    Returns:
        The filtered dataframe as a `~tfs.TfsDataFrame`.
    """
    # rename measurement column to key
    column_map = {c[0]: c for c in [REAL, IMAG, AMPLITUDE, PHASE]}  # only REAL and IMAG implemented in responses so far
    meas_col = column_map[col[-1]]
    meas.columns = meas.columns.str.replace(meas_col, col)
    return _get_filtered_generic(col, meas, model, opt)


def _get_errorbased_weights(key: str, weights, errors):
    # TODO case without errors used may corrupt the correction (typical error != 1)
    w2 = stats.weights_from_errors(errors)
    if w2 is None:
        LOG.warn(
            f"Weights will not be based on errors for '{key}'"
            f", zeros of NaNs were found. Maybe don't use --errorbars."
        )
        return weights
    return weights * np.sqrt(w2)


# Response Matrix Filter -------------------------------------------------------


def filter_response_index(response: Dict, measurement: Dict, keys: Sequence[str]):
    """Filters the index of the response matrices `response` by the respective entries in `measurement`."""
    # rename MU to PHASE as we create a PHASE-Response afterwards
    # easier to do here, than to check eveywhere below. (jdilly)
    _rename_phase_advance(response)

    not_in_response = [key for key in keys if key not in response]
    if len(not_in_response) > 0:
        raise KeyError(
            f"The following optical parameters are not present in current"
            f"response matrix: {not_in_response}"
        )

    filters = _get_response_filters()
    new_response = {}
    for key in keys:
        new_response[key] = filters[key](response[key], measurement[key])
    return new_response


def _get_response_filters() -> Dict[str, Callable]:
    """
    Returns a dict with the respective `_get_*_response` functions that defaults
    to `_get_generic_response`.
    """
    return defaultdict(
        lambda: _get_generic_response,
        {f"{PHASE}X": _get_phase_response, f"{PHASE}Y": _get_phase_response},
    )


def _get_generic_response(resp: pd.DataFrame, meas: pd.DataFrame) -> pd.DataFrame:
    return resp.loc[meas.index.to_numpy(), :]


def _get_phase_response(resp: pd.DataFrame, meas: pd.DataFrame) -> pd.DataFrame:
    """Creates response for PHASE, not MU."""
    phase1 = resp.loc[meas.index.to_numpy(), :]
    phase2 = resp.loc[meas.loc[:, NAME2].to_numpy(), :]
    return -phase1.sub(phase2.to_numpy())  # phs2-phs1 but with idx of phs1


def _get_smallest_data_mask(data, portion: float = 0.95) -> np.ndarray:
    if not 0 <= portion <= 1:
        raise ValueError("Portion of data has to be between 0 and 1")
    b = int(len(data) * portion)
    mask = np.ones_like(data, dtype=bool)
    mask[np.argpartition(data, b)[b:]] = False
    return mask


def _rename_phase_advance(response):
    """Renames MU to PHASE inplace."""
    for plane in PLANES:
        try:
            response[f"{PHASE}{plane}"] = response.pop(f"{PHASE_ADV}{plane}")
        except KeyError:
            pass



"""
Iota
----

Accelerator-Class for the ``Iota`` machine.

Model Creation Keyword Args:
    *--Optional--*

    - **dpp** *(float)*:

        Deltap/p to use.

        default: ``0.0``


    - **driven_excitation** *(str)*:

        Denotes driven excitation by `AC-dipole` (acd) or by `ADT` (adt)

        choices: ``('acd', 'adt')``


    - **drv_tunes** *(float)*:

        Driven tunes without integer part.


    - **energy** *(float)*:

        Energy in **Tev**.


    - **model_dir** *(str)*:

        Path to model directory; loads tunes and excitation from model!


    - **modifiers** *(str)*:

        Path to the optics file to use (modifiers file).


    - **nat_tunes** *(float)*:

        Natural tunes without integer part.


    - **particle** *(str)*:

        Particle type.

        choices: ``('p', 'e')``


    - **xing**:

        If True, x-ing angles will be applied to model

        action: ``store_true``
"""
import logging
import os

from generic_parser import EntryPoint

from omc3.model.accelerators.accelerator import AccElementTypes, Accelerator

LOGGER = logging.getLogger(__name__)
CURRENT_DIR = os.path.dirname(__file__)


class Iota(Accelerator):
    NAME = "iota"
    RE_DICT = {AccElementTypes.BPMS: r"IBPM*",
               AccElementTypes.MAGNETS: r"Q*",
               AccElementTypes.ARC_BPMS: r"IBPM*"}
    BPM_INITIAL = 'I'

    @staticmethod
    def get_parameters():
        params = super(Iota, Iota).get_parameters()
        params.add_parameter(name="particle", type=str, choices=('p', 'e'), help="Particle type.")
        return params

    def __init__(self, *args, **kwargs):
        parser = EntryPoint(self.get_parameters(), strict=True)
        opt = parser.parse(*args, **kwargs)
        super().__init__(opt)
        self.particle = opt.particle

    @classmethod
    def verify_object(self):
        pass  # TODO



"""
TbT Converter
-------------

Top-level script to convert turn-by-turn files from various formats to ``LHC`` binary SDDS files.
Optionally, it can replicate files with added noise.
"""
import copy
from datetime import datetime
from pathlib import Path
from typing import Sequence, Union

from generic_parser.entrypoint_parser import (
    EntryPointParameters,
    entrypoint,
    save_options_to_config,
)

import turn_by_turn as tbt
from omc3.definitions import formats
from omc3.utils import iotools, logging_tools

LOGGER = logging_tools.get_logger(__name__)

DEFAULT_CONFIG_FILENAME = "converter_{time:s}.ini"


def converter_params():
    params = EntryPointParameters()
    params.add_parameter(name="files", required=True, nargs="+", help="TbT files to analyse")
    params.add_parameter(name="outputdir", required=True, help="Output directory.")
    params.add_parameter(
        name="tbt_datatype",
        type=str,
        default="lhc",
        choices=list(tbt.io.TBT_MODULES.keys()),
        help="Choose the datatype from which to import. ",
    )
    params.add_parameter(
        name="output_datatype",
        type=str,
        default="lhc",
        choices=list(tbt.io.WRITERS),
        help="Choose the datatype to export. ",
    )
    params.add_parameter(
        name="realizations", type=int, default=1, help="Number of copies with added noise"
    )
    params.add_parameter(name="noise_levels", nargs="+", help="Sigma of added Gaussian noise")
    params.add_parameter(
        name="use_average",
        action="store_true",
        help="If set, returned sdds only contains the average over all particle/bunches.",
    )
    params.add_parameter(
        name="drop_elements",
        nargs="+",
        help="Names of elements to drop from the input file during conversion",
    )
    return params


@entrypoint(converter_params(), strict=True)
def converter_entrypoint(opt):
    """
    Converts turn-by-turn files from various formats to ``LHC`` binary SDDS files.
    Optionally can replicate files with added noise.

    Converter Kwargs:
      - **files**: TbT files to convert

        Flags: **--files**
        Required: ``True``
      - **outputdir**: Output directory.

        Flags: **--outputdir**
        Required: ``True``
      - **tbt_datatype** *(str)*: Choose datatype from which to import (e.g LHC binary SDDS).

        Flags: **--tbt_datatype**
        Default: ``lhc``
      - **output_datatype** *(str)*: Choose the datatype to export.

        Flags: **--output_datatype**
        Default: ``lhc``
      - **realizations** *(int)*: Number of copies with added noise.

        Flags: **--realizations**
        Default: ``1``
      - **noise_levels** *(float)*: Sigma of added Gaussian noise.

        Flags: **--noise_levels**
        Default: ``None``
      - **use_average** *(bool)*: If set, returned sdds only contains the average over all particle/bunches.

        Flags: **--use_average**
        Default: ``False``

      - **drop_elements**: Names of elements to drop from the input file during conversion.

        Flags: **--drop_elements**
        Default: ``None``
    """
    if opt.realizations < 1:
        raise ValueError("Number of realizations lower than 1.")
    iotools.create_dirs(opt.outputdir)
    save_options_to_config(
        str(
            Path(opt.outputdir)
            / DEFAULT_CONFIG_FILENAME.format(time=datetime.utcnow().strftime(formats.TIME))
        ),
        dict(sorted(opt.items())),
    )
    _read_and_write_files(opt)


def _read_and_write_files(opt):
    for input_file in opt.files:
        tbt_data = tbt.read_tbt(input_file, datatype=opt.tbt_datatype)
        if opt.drop_elements:
            tbt_data = _drop_elements(tbt_data, opt.drop_elements)
        if opt.use_average:
            tbt_data = tbt.utils.generate_average_tbtdata(tbt_data)
        for i in range(opt.realizations):
            suffix = f"_r{i}" if opt.realizations > 1 else ""
            if opt.noise_levels is None:
                tbt.write(
                    Path(opt.outputdir) / f"{_file_name_without_sdds(input_file)}{suffix}",
                    tbt_data=tbt_data,
                )
            else:
                for noise_level in opt.noise_levels:
                    tbt.write(
                        Path(opt.outputdir) / f"{_file_name_without_sdds(input_file)}_n{noise_level}{suffix}",
                        datatype=opt.output_datatype,
                        tbt_data=tbt_data,
                        noise=float(noise_level),
                    )


def _drop_elements(tbt_data: tbt.TbtData, elements_to_drop: Sequence[str]) -> tbt.TbtData:
    """
    Drops the provided elements from the matrices in the provided TbtData object.
    For any element not found in the matrices, a warning is logged and the element is skipped.

    Args:
        tbt_data (tbt.TbtData): a TbTData object from loading your turn-by-turn data from disk.
        elements_to_drop (Sequence[str]): list of elements to drop.

    Returns:
        A copied version of the provided TbtData object with the relevant element dropped from the matrices.
    """
    copied_data: tbt.TbtData = copy.deepcopy(tbt_data)
    LOGGER.info(f"Dropping the following unwanted elements: {', '.join(elements_to_drop)}")
    for element in elements_to_drop:
        LOGGER.debug(f"Dropping element '{element}'")
        try:
            for transverse_data in copied_data.matrices:
                transverse_data.X = transverse_data.X.drop(element)
                transverse_data.Y = transverse_data.Y.drop(element)
        except (KeyError, AttributeError):
            LOGGER.warning(f"Element '{element}' could not be found, skipped")
    return copied_data


def _file_name_without_sdds(filename: Union[str, Path]) -> str:
    """
    Returns the file name without suffix if the suffix is **.sdds**, else the filename.
    Previous elements in the file path are discarded.
    """
    filename = Path(filename)
    return filename.stem if filename.suffix == ".sdds" else filename.name


if __name__ == "__main__":
    converter_entrypoint()



"""
Time Tools
-------------------------

Provides tools to handle times more easily, in particular to switch easily between local time
and UTC time.
"""
from datetime import datetime, timedelta
import dateutil.tz as tz

from omc3.definitions.formats import TIME


# Datetime Conversions #########################################################


def utc_now():
    """Get UTC now as time."""
    return datetime.now(tz.tzutc())


def get_cern_timezone():
    """Get time zone for cern measurement data."""
    return tz.gettz('Europe/Zurich')


def get_cern_time_format():
    """Default time format."""
    return TIME  # TODO maybe only here?


def get_readable_time_format():
    """Human readable time format."""
    return "%Y-%m-%d %H:%M:%S.%f"


def local_to_utc(dt_local, timezone):
    """Convert local datetime object to utc datetime object."""
    check_tz(dt_local, timezone)

    dt_utc = dt_local.astimezone(tz.tzutc())
    return dt_utc


def utc_to_local(dt_utc, timezone):
    """Convert UTC datetime object to local datetime object."""
    check_tz(dt_utc, tz.tzutc())

    dt_local = dt_utc.astimezone(timezone)
    return dt_local


def local_string_to_utc(local_string, timezone):
    """Converts a time string in local time to UTC time."""
    dt = datetime.strptime(local_string, get_readable_time_format())
    dt = dt.replace(tzinfo=timezone)
    return local_to_utc(dt, timezone)


def utc_string_to_utc(utc_string):
    """Convert a time string in utc to a UTC datetime object."""
    dt = datetime.strptime(utc_string, get_readable_time_format())
    dt = dt.replace(tzinfo=tz.tzutc())
    return dt


def cern_utc_string_to_utc(utc_string):
    """Convert a time string in cern-utc to a utc datetime object."""
    dt = datetime.strptime(utc_string, get_cern_time_format())
    dt = dt.replace(tzinfo=tz.tzutc())
    return dt


def check_tz(localized_dt, timezone):
    """Checks if timezone is correct."""
    if localized_dt.tzinfo is None or localized_dt.tzinfo.utcoffset(localized_dt) is None:
        raise AssertionError("Datetime object needs to be localized!")

    if not localized_dt.tzname() == datetime.now(timezone).tzname():
        raise AssertionError(
            f"Datetime Timezone should be '{timezone}' "
            f"but was '{localized_dt.tzinfo}'"
        )


# AccDatetime Classes ##########################################################


class AccDatetime(datetime):
    """
    Wrapper for a datetime object to easily convert between local and UTC time as well as give
    different presentations.
    """
    _LOCAL_TIMEZONE = None
    nanosecond = 0  # used in by pandas to make indices

    def __new__(cls, *args, **kwargs):
        if cls._LOCAL_TIMEZONE is None:
            raise NotImplementedError("Do not use the AccDatetime class, "
                                      "but one of its children.")
        if len(args) == 1 and isinstance(args[0], datetime):
            dt = args[0]
        else:
            dt = datetime.__new__(cls, *args, **kwargs)

        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=tz.tzutc())

        return datetime.__new__(cls, dt.year, dt.month, dt.day,
                                dt.hour, dt.minute, dt.second, dt.microsecond,
                                tzinfo=dt.tzinfo)

    @property
    def datetime(self):
        """
        Return normal datetime object (in case ducktyping does not work. Looking at you, mpl!).
        """
        return datetime(self.year, self.month, self.day,
                        self.hour, self.minute, self.second, self.microsecond,
                        tzinfo=tz.tzutc())

    @property
    def local_timezone(self):
        """Get local timezone."""
        return self._LOCAL_TIMEZONE

    @property
    def utc(self):
        """Get UTC datetime object."""
        return self

    @property
    def local(self):
        """Get local datetime object."""
        return self.astimezone(self._LOCAL_TIMEZONE)

    @property
    def local_string(self):
        """Get local time as string."""
        return self.local.strftime(get_readable_time_format())

    @property
    def utc_string(self):
        """Get utc time as string."""
        return self.strftime(get_readable_time_format())

    def cern_utc_string(self):
        """Get utc time as string (CERN format)."""
        return self.strftime(get_cern_time_format())

    def add(self, **kwargs):
        """Add timedelta and return as new object."""
        return self.__class__(self + timedelta(**kwargs))

    def sub(self, **kwargs):
        """Subtract timedelta and return as new object."""
        return self.__class__(self - timedelta(**kwargs))

    def __add__(self, td):
        """Add timedelta and return as new object."""
        return self.__class__(super().__add__(td))

    def __sub__(self, other):
        """Subtract timedelta or datetime/AccDatetime and return as new object."""
        result = super().__sub__(other)
        if isinstance(result, datetime):
            return self.__class__(result)
        if isinstance(result, timedelta):
                return result
        return NotImplemented

    @classmethod
    def from_local_string(cls, s):
        """Create `AccDatetime` object from datetime in local time string."""
        return cls(local_string_to_utc(s, cls._LOCAL_TIMEZONE))

    @classmethod
    def from_utc_string(cls, s):
        """Create `AccDatetime` object from datetime in utc string."""
        return cls(utc_string_to_utc(s))

    @classmethod
    def from_cern_utc_string(cls, s):
        """Create `AccDatetime` object from datetime in utc string."""
        return cls(cern_utc_string_to_utc(s))

    @classmethod
    def from_local(cls, dt):
        """Create AccDatetime object from datetime in local time."""
        return cls(local_to_utc(dt, cls._LOCAL_TIMEZONE))

    @classmethod
    def from_utc(cls, dt):
        """Create `AccDatetime` object from datetime in utc."""
        return cls(dt)

    @classmethod
    def from_timestamp(cls, ts):
        """Create `AccDatetime` object from timestamp."""
        return cls(datetime.utcfromtimestamp(ts))

    @classmethod
    def now(cls):
        """Create `AccDatetime` object at now."""
        return cls(utc_now())


class CERNDatetime(AccDatetime):
    """`AccDatetime` class for all accelerators at CERN."""
    _LOCAL_TIMEZONE = get_cern_timezone()


AcceleratorDatetime = {
    'lhc': CERNDatetime,
    'ps': CERNDatetime,
    'sps': CERNDatetime,
}
"""dict: Accelerator name to AccDatetime mapping. """



"""
Plot BBQ
--------

Provides the plotting function for the extracted and cleaned BBQ data from timber.

**Arguments:**

*--Required--*

- **input**:

    BBQ data as data frame or tfs file.


*--Optional--*

- **interval** *(float)*:

    x_axis interval that was used in calculations.


- **kick**:

    Kick file as data frame or tfs file.


- **manual_style** *(DictAsString)*:

    Additional style rcParameters which update the set of predefined ones.

    default: ``{}``


- **output** *(str)*:

    Save figure to this location.


- **plot_styles** *(UnionPathStr)*:

    Which plotting styles to use, either from plotting.styles.*.mplstyles
    or default mpl.

    default: ``['standard', 'bbq']``


- **show**:

    Show plot.

    action: ``store_true``


- **two_plots**:

    Plot two axis into the figure.

    action: ``store_true``


- **x_lim** *(float)*:

    X-Axis limits. (yyyy-mm-dd HH:mm:ss.mmm)


- **y_lim** *(float)*:

    Y-Axis limits.
    

"""
from collections import OrderedDict
from contextlib import suppress
from pathlib import Path

import matplotlib.dates as mdates
import numpy as np
from generic_parser import entrypoint, EntryPointParameters
from generic_parser.entry_datatypes import DictAsString
from generic_parser.entrypoint_parser import save_options_to_config
from matplotlib import pyplot as plt
from matplotlib.ticker import FormatStrFormatter
from pandas.plotting import register_matplotlib_converters

from omc3 import amplitude_detuning_analysis as ad_ana
from omc3.definitions import formats
from omc3.definitions.constants import PLANES
from omc3.plotting.utils import colors as pcolors, style as pstyle
from omc3.tune_analysis import kick_file_modifiers as kick_mod
from omc3.tune_analysis.constants import (get_mav_window_header, get_used_in_mav_col,
                                          get_bbq_col, get_mav_col)
from omc3.utils import logging_tools
from omc3.utils.iotools import UnionPathStr, PathOrStr, PathOrStrOrDataFrame

LOG = logging_tools.get_logger(__name__)

# Registering converters for datetime plotting as pandas won't do it for us automatically anymore
register_matplotlib_converters()


def get_params():
    params = EntryPointParameters()
    params.add_parameter(
        name="input",
        help="BBQ data as data frame or tfs file.",
        required=True,
        type=PathOrStrOrDataFrame
    )
    params.add_parameter(
        name="kick",
        help="Kick file as data frame or tfs file.",
        type=PathOrStrOrDataFrame
    )
    params.add_parameter(
        name="output",
        help="Save figure to this location.",
        type=PathOrStr,
    )
    params.add_parameter(
        name="show",
        help="Show plot.",
        action="store_true"
    )
    params.add_parameter(
        name="x_lim",
        help="X-Axis limits. (yyyy-mm-dd HH:mm:ss.mmm)",
        type=float,
        nargs=2,
    )
    params.add_parameter(
        name="y_lim",
        help="Y-Axis limits.",
        type=float,
        nargs=2,
    )
    params.add_parameter(
        name="interval",
        help="x_axis interval that was used in calculations.",
        type=float,
        nargs=2,
    )
    params.add_parameter(
        name="two_plots",
        help="Plot two axis into the figure.",
        action="store_true",
    )
    params.add_parameter(
        name="plot_styles",
        type=UnionPathStr,
        nargs="+",
        default=['standard', 'bbq'],
        help='Which plotting styles to use, either from plotting.styles.*.mplstyles or default mpl.'
    )
    params.add_parameter(
        name="manual_style",
        type=DictAsString,
        default={},
        help='Additional style rcParameters which update the set of predefined ones.'
    )
    return params


@entrypoint(get_params(), strict=True)
def main(opt):
    """Plot BBQ wrapper."""
    LOG.info("Plotting BBQ.")
    _save_options(opt)
    pstyle.set_style(opt.pop("plot_styles"), opt.pop("manual_style"))

    bbq_df = kick_mod.read_timed_dataframe(opt.input) if isinstance(opt.input, (Path, str)) else opt.input
    opt.pop("input")

    if opt.kick is not None:
        if opt.interval is not None:
            raise ValueError("interval and kick-file given. Both are used for the same purpose. Please only use one.")

        window = 0  # not too important, bars will then indicate first and last kick directly
        with suppress(KeyError):
            window = bbq_df.headers[get_mav_window_header()]

        kick_df = kick_mod.read_timed_dataframe(opt.kick) if isinstance(opt.kick, (Path, str)) else opt.kick
        opt.interval = ad_ana.get_approx_bbq_interval(bbq_df, kick_df.index, window)
        bbq_df = bbq_df.loc[opt.interval[0]:opt.interval[1]]
    opt.pop("kick")

    show = opt.pop("show")
    out = opt.pop("output")

    fig = _plot_bbq_data(bbq_df, **opt)

    if show:
        plt.show()
    if out:
        fig.savefig(out)

    return fig


def _plot_bbq_data(bbq_df, interval=None, x_lim=None, y_lim=None, two_plots=False):
    """
    Plot BBQ data.

    Args:
        bbq_df: BBQ Dataframe with moving average columns.
        interval: start and end time of used interval, will be marked with red bars.
        x_lim: x limits (time).
        y_lim: y limits (tune).
        output: Path to the output file.
        show: Shows plot if ``True``.
        two_plots: Plots each tune in it's own axes if ``True``.

    Returns:
        Plotted figure.
    """
    LOG.debug("Plotting BBQ data.")
    fig, axs = plt.subplots(1+two_plots, 1)

    if not two_plots:
        axs = [axs, axs]

    handles = [None] * (3 * len(PLANES))
    for idx, plane in enumerate(PLANES):
        color = pcolors.get_mpl_color(idx)
        mask = np.array(bbq_df[get_used_in_mav_col(plane)], dtype=bool)

        # plot and save handles for nicer legend
        handles[idx] = axs[idx].plot([i.datetime for i in bbq_df.index],
                                    bbq_df[get_bbq_col(plane)],
                                    color=pcolors.change_color_brightness(color, .4),
                                    marker="o", markerfacecolor="None",
                                    label="$Q_{:s}$".format(plane.lower(),)
                                    )[0]
        filtered_data = bbq_df.loc[mask, get_bbq_col(plane)].dropna()
        handles[len(PLANES)+idx] = axs[idx].plot(filtered_data.index, filtered_data.to_numpy(),
                                                color=pcolors.change_color_brightness(color, .7),
                                                marker=".",
                                                label="filtered".format(plane.lower())
                                                )[0]
        handles[2*len(PLANES)+idx] = axs[idx].plot(bbq_df.index, bbq_df[get_mav_col(plane)],
                                                  color=color,
                                                  linestyle="-",
                                                  label="moving av.".format(plane.lower())
                                                  )[0]

        if (y_lim is None or y_lim[0] is None) and two_plots:
            axs[idx].set_ylim(bottom=min(bbq_df.loc[mask, get_bbq_col(plane)]))

        if (y_lim is None or y_lim[1] is None) and two_plots:
            axs[idx].set_ylim(top=max(bbq_df.loc[mask, get_bbq_col(plane)]))

    # things to add/do only once if there is only one plot
    for idx in range(1+two_plots):
        if interval:
            axs[idx].axvline(x=interval[0], color="red")
            axs[idx].axvline(x=interval[1], color="red")

        if two_plots:
            axs[idx].set_ylabel("$Q_{:s}$".format(PLANES[idx]))
        else:
            axs[idx].set_ylabel('Tune')

        if y_lim is not None:
            axs[idx].set_ylim(y_lim)
        axs[idx].yaxis.set_major_formatter(FormatStrFormatter('%.5f'))

        if x_lim is not None:
            axs[idx].set_xlim(x_lim)
        axs[idx].set_xlabel('Time')
        axs[idx].xaxis.set_major_formatter(mdates.DateFormatter('%H:%M:%S'))

        if idx:
            # don't show labels on upper plot (if two plots)
            # use the visibility to allow cursor x-position to be shown
            axs[idx].tick_params(labelbottom=False)
            axs[idx].xaxis.get_label().set_visible(False)

        if not two_plots or idx:
            # reorder legend
            axs[idx].legend(handles, [h.get_label() for h in handles],
                           loc='lower right', bbox_to_anchor=(1.0, 1.01), ncol=3,)
    return fig


def _save_options(opt):
    if opt.output:
        out_path = Path(opt.output).parent
        out_path.mkdir(exist_ok=True, parents=True)
        save_options_to_config(str(out_path / formats.get_config_filename(__file__)),
                               OrderedDict(sorted(opt.items()))
                               )

# Script Mode ------------------------------------------------------------------


if __name__ == '__main__':
    main()



"""
LHC
---

Accelerator-Class for the ``LHC`` collider.

Model Creation Keyword Args:
    *--Required--*

    - **beam** *(int)*:

        Beam to use.

        choices: ``(1, 2)``


    - **year** *(str)*:

        Year of the optics (or hllhc1.3).

        choices: ``('2012', '2015', '2016', '2017', '2018', '2022', 'hllhc1.3')``


    *--Optional--*

    - **ats**:

        Marks ATS optics

        action: ``store_true``


    - **dpp** *(float)*:

        Delta p/p to use.

        default: ``0.0``


    - **driven_excitation** *(str)*:

        Denotes driven excitation by AC-dipole (acd) or by ADT (adt)

        choices: ``('acd', 'adt')``


    - **drv_tunes** *(float)*:

        Driven tunes without integer part.


    - **energy** *(float)*:

        Energy in Tev.


    - **model_dir** *(str)*:

        Path to model directory; loads tunes and excitation from model!


    - **modifiers** *(str)*:

        Path to the optics file to use (modifiers file).


    - **nat_tunes** *(float)*:

        Natural tunes without integer part.


    - **xing**:

        If True, x-ing angles will be applied to model

        action: ``store_true``


"""
import json
from collections import OrderedDict
from pathlib import Path
from typing import Dict, Iterator, List, Sequence, Tuple

import tfs
from generic_parser import EntryPoint

from omc3.model.accelerators.accelerator import (
    AccElementTypes,
    Accelerator,
    AcceleratorDefinitionError,
    AccExcitationMode,
)
from omc3.model.constants import (
    B2_ERRORS_TFS,
    B2_SETTINGS_MADX,
    GENERAL_MACROS,
    LHC_MACROS,
    LHC_MACROS_RUN3,
    MACROS_DIR,
    MODIFIER_TAG,
)
from omc3.utils import logging_tools

LOGGER = logging_tools.get_logger(__name__)
CURRENT_DIR = Path(__file__).parent
LHC_DIR = CURRENT_DIR / "lhc"


class Lhc(Accelerator):
    """
    Parent Class for LHC-types.
    """

    NAME = "lhc"
    RE_DICT: Dict[str, str] = {
        AccElementTypes.BPMS: r"BPM",
        AccElementTypes.MAGNETS: r"M",
        AccElementTypes.ARC_BPMS: r"BPM.*\.0*(1[5-9]|[2-9]\d|[1-9]\d{2,})[RL]",
    }  # bpms > 14 L or R of IP

    LHC_IPS = ("1", "2", "5", "8")
    NORMAL_IP_BPMS = "BPMSW.1{side}{ip}.B{beam}"
    DOROS_IP_BPMS = "LHC.BPM.1{side}{ip}.B{beam}_DOROS"

    @staticmethod
    def get_parameters():
        params = super(Lhc, Lhc).get_parameters()
        params.add_parameter(
            name="beam", type=int, choices=(1, 2), required=True, help="Beam to use."
        )
        params.add_parameter(
            name="year",
            type=str,
            required=True,
            choices=("2012", "2015", "2016", "2017", "2018", "2022", "hllhc1.3"),
            help="Year of the optics (or hllhc1.x version).",
        )
        params.add_parameter(
            name="ats",
            action="store_true",
            help="Force use of ATS macros and knobs for years which are not ATS by default.",
            )
        return params

    def __init__(self, *args, **kwargs):
        parser = EntryPoint(self.get_parameters(), strict=True)
        opt = parser.parse(*args, **kwargs)
        super().__init__(opt)
        self.correctors_dir = "2012"
        self.year = opt.year
        self.ats = opt.ats
        if self.year == "hllhc1.3":
            self.correctors_dir = "hllhc1.3"
        self.beam = opt.beam
        beam_to_beam_direction = {1: 1, 2: -1}
        self.beam_direction = beam_to_beam_direction[self.beam]
        self.verify_object()

    def verify_object(self) -> None:  # TODO: Maybe more checks?
        """
        Verifies if everything is defined which should be defined.
        Will Raise an ``AcceleratorDefinitionError`` if one of the checks is invalid.
        """
        LOGGER.debug("Accelerator class verification")
        _ = self.beam

        if self.model_dir is None and self.xing is None:
            raise AcceleratorDefinitionError("Crossing on or off not set.")

        if self.excitation is None:
            raise AcceleratorDefinitionError("Excitation mode not set.")
        if (self.excitation != AccExcitationMode.FREE) and (self.drv_tunes is None):
            raise AcceleratorDefinitionError("An excitation mode was given but driven tunes are not set.")

        # TODO: write more output prints
        LOGGER.debug("... verification passed. \nSome information about the accelerator:")
        LOGGER.debug(f"Class name       {self.__class__.__name__}")
        LOGGER.debug(f"Beam             {self.beam}")
        LOGGER.debug(f"Beam direction   {self.beam_direction}")
        if self.modifiers:
            LOGGER.debug(f"Modifiers        {', '.join([str(m) for m in self.modifiers])}")

    @property
    def beam(self) -> int:
        if self._beam is None:
            raise AcceleratorDefinitionError(
                "The accelerator definition is incomplete, beam "
                "has to be specified (--beam option missing?)."
            )
        return self._beam

    @beam.setter
    def beam(self, value) -> None:
        if value not in (1, 2):
            raise AcceleratorDefinitionError("Beam parameter has to be one of (1, 2)")
        self._beam = value

    @staticmethod
    def get_lhc_error_dir() -> Path:
        return LHC_DIR / "systematic_errors"

    def get_variables(self, frm=None, to=None, classes=None):
        correctors_dir = LHC_DIR / "2012" / "correctors"
        all_corrs = _merge_jsons(
            correctors_dir / f"correctors_b{self.beam}" / "beta_correctors.json",
            correctors_dir / f"correctors_b{self.beam}" / "coupling_correctors.json",
            self._get_triplet_correctors_file(),
        )
        my_classes = classes
        if my_classes is None:
            my_classes = all_corrs.keys()
        vars_by_class = set(
            _flatten_list([all_corrs[corr_cls] for corr_cls in my_classes if corr_cls in all_corrs])
        )
        if frm is None and to is None:
            return list(vars_by_class)
        elems_matrix = tfs.read(self._get_corrector_elems()).sort_values("S")
        if frm is not None and to is not None:
            if frm > to:
                elems_matrix = elems_matrix[(elems_matrix.S >= frm) | (elems_matrix.S <= to)]
            else:
                elems_matrix = elems_matrix[(elems_matrix.S >= frm) & (elems_matrix.S <= to)]
        elif frm is not None:
            elems_matrix = elems_matrix[elems_matrix.S >= frm]
        elif to is not None:
            elems_matrix = elems_matrix[elems_matrix.S <= to]

        vars_by_position = _remove_dups_keep_order(
            _flatten_list([raw_vars.split(",") for raw_vars in elems_matrix.loc[:, "VARS"]])
        )
        return _list_intersect_keep_order(vars_by_position, vars_by_class)

    def get_ips(self) -> Iterator[Tuple[str]]:
        """
        Returns an iterable with this accelerator's IPs.

        Returns:
            An iterator returning `tuples` with:
                (``ip name``, ``left BPM name``, ``right BPM name``)
        """
        for ip in Lhc.LHC_IPS:
            yield (
                f"IP{ip}",
                Lhc.NORMAL_IP_BPMS.format(side="L", ip=ip, beam=self.beam),
                Lhc.NORMAL_IP_BPMS.format(side="R", ip=ip, beam=self.beam),
            )
            yield (
                f"IP{ip}_DOROS",
                Lhc.DOROS_IP_BPMS.format(side="L", ip=ip, beam=self.beam),
                Lhc.DOROS_IP_BPMS.format(side="R", ip=ip, beam=self.beam),
            )

    def log_status(self) -> None:
        LOGGER.info(f"  model dir = {self.model_dir}")
        LOGGER.info(f"Natural Tune X      [{self.nat_tunes[0]:10.3f}]")
        LOGGER.info(f"Natural Tune Y      [{self.nat_tunes[1]:10.3f}]")
        LOGGER.info(
            f"Best Knowledge Model     "
            f"[{'NO' if self.model_best_knowledge is None else 'OK':>10s}]"
        )

        if self.excitation == AccExcitationMode.FREE:
            LOGGER.info(f"Excitation          [{'NO':>10s}]")
            return
        LOGGER.info(
            f"Excitation          "
            f"[{'ACD' if self.excitation == AccExcitationMode.ACD else 'ADT':>10s}]"
        )
        LOGGER.info(f"> Driven Tune X     [{self.drv_tunes[0]:10.3f}]")
        LOGGER.info(f"> Driven Tune Y     [{self.drv_tunes[1]:10.3f}]")

    def load_main_seq_madx(self) -> str:
        try:
            return _get_call_main_for_year(self.year)
        except AttributeError:
            raise AcceleratorDefinitionError(
                "The accelerator definition is incomplete, mode "
                "has to be specified (--lhcmode option missing?)."
            )

    # Private Methods ##########################################################

    def _get_triplet_correctors_file(self) -> Path:
        correctors_dir = LHC_DIR / self.correctors_dir / "correctors"
        return correctors_dir / "triplet_correctors.json"

    def _get_corrector_elems(self) -> Path:
        correctors_dir = LHC_DIR / self.correctors_dir / "correctors"
        return correctors_dir / f"corrector_elems_b{self.beam}.tfs"

    def get_exciter_bpm(self, plane: str, commonbpms: List[str]):
        beam = self.beam
        adt = "H.C" if plane == "X" else "V.B"
        l_r = "L" if (beam == 1 != plane == "Y") else "R"
        a_b = "B" if beam == 1 else "A"
        if self.excitation == AccExcitationMode.ACD:
            try:
                return (
                    _is_one_of_in([f"BPMY{a_b}.6L4.B{beam}", f"BPM.7L4.B{beam}"], commonbpms),
                    f"MKQA.6L4.B{beam}",
                )
            except KeyError as e:
                raise KeyError("AC-Dipole BPM not found in the common BPMs. Maybe cleaned?") from e
        if self.excitation == AccExcitationMode.ADT:
            try:
                return (
                    _is_one_of_in([f"BPMWA.B5{l_r}4.B{beam}", f"BPMWA.A5{l_r}4.B{beam}"], commonbpms),
                    f"ADTK{adt}5{l_r}4.B{beam}",
                )
            except KeyError as e:
                raise KeyError("ADT BPM not found in the common BPMs. Maybe cleaned?") from e
        return None

    def important_phase_advances(self) -> List[List[str]]:
        if self.beam == 2:
            return [["MKD.O5R6.B2", "TCTPH.4R1.B2"], ["MKD.O5R6.B2", "TCTPH.4R5.B2"]]
        if self.beam == 1:
            return [["MKD.O5L6.B1", "TCTPH.4L1.B1"], ["MKD.O5L6.B1", "TCTPH.4L5.B1"]]

    def get_synch_BPMs(self, index):
        # expect passing index.to_numpy()
        if self.beam == 1:
            return [i in index for i in self.model.loc["BPMSW.33L2.B1":].index]
        elif self.beam == 2:
            return [i in index for i in self.model.loc["BPMSW.33R8.B2":].index]

    def _get_madx_script_info_comments(self) -> str:
        info_comments = (
            f'title, "LHC Model created by omc3";\n'
            f"! Model directory: {Path(self.model_dir).absolute()}\n"
            f"! Natural Tune X         [{self.nat_tunes[0]:10.3f}]\n"
            f"! Natural Tune Y         [{self.nat_tunes[1]:10.3f}]\n"
            f"! Best Knowledge:        [{'NO' if self.model_best_knowledge is None else 'OK':>10s}]\n"
        )
        if self.excitation == AccExcitationMode.FREE:
            info_comments += f"! Excitation             [{'NO':>10s}]\n\n"
            return info_comments
        else:
            info_comments += (
                f"! Excitation             [{'ACD' if self.excitation == AccExcitationMode.ACD else 'ADT':>10s}]\n"
                f"! > Driven Tune X        [{self.drv_tunes[0]:10.3f}]\n"
                f"! > Driven Tune Y        [{self.drv_tunes[1]:10.3f}]\n\n"
            )
        return info_comments

    def get_base_madx_script(self, best_knowledge: bool = False) -> str:
        ats_md = False
        high_beta = False
        madx_script = (
            f"{self._get_madx_script_info_comments()}"
            f"! ----- Calling Sequence and Optics -----\n"
            f"call, file = '{self.model_dir / MACROS_DIR / GENERAL_MACROS}';\n"
            f"call, file = '{self.model_dir / MACROS_DIR / LHC_MACROS}';\n"
            )
        if self._uses_run3_macros():
            LOGGER.debug("According to the optics year, Run 3 versions of the macros will be used")
            madx_script += (
                f"call, file = '{self.model_dir / MACROS_DIR / LHC_MACROS_RUN3}';\n"
            )

        madx_script += (
            f"{self.load_main_seq_madx()}\n"
            f"exec, define_nominal_beams();\n"
        )
        madx_script += "".join(
            f"call, file = '{self.model_dir / modifier}'; {MODIFIER_TAG}\n"
            for modifier in self.modifiers
        )
        madx_script += (
            f"\n! ----- Defining Configuration Specifics -----\n"
            f"exec, cycle_sequences();\n"
            f"xing_angles = {'1' if self.xing else '0'};\n"
            f"if(xing_angles==1){{\n"
            f"    exec, set_crossing_scheme_ON();\n"
            f"}}else{{\n"
            f"    exec, set_default_crossing_scheme();\n"
            f"}}\n"
            f"use, sequence = LHCB{self.beam};\n"
            f"option, echo;\n"
        )
        if best_knowledge:
            # madx_script += f"exec, load_average_error_table({self.energy}, {self.beam});\n"
            madx_script += (
                f"\n! ----- For Best Knowledge Model -----\n"
                f"readmytable, file = '{self.model_dir / B2_ERRORS_TFS}', table=errtab;\n"
                f"seterr, table=errtab;\n"
                f"call, file = '{self.model_dir / B2_SETTINGS_MADX}';\n"
            )
        if high_beta:
            madx_script += "exec, high_beta_matcher();\n"

        madx_script += f"\n! ----- Matching Knobs and Output Files -----\n"
        if self._uses_ats_knobs():
            LOGGER.debug("According to the optics year or the --ats flag being provided, ATS macros and knobs will be used")
            madx_script += f"exec, match_tunes_ats({self.nat_tunes[0]}, {self.nat_tunes[1]}, {self.beam});\n"
            madx_script += f"exec, coupling_knob_ats({self.beam});\n"
        else:
            madx_script += f"exec, match_tunes({self.nat_tunes[0]}, {self.nat_tunes[1]}, {self.beam});\n"
            madx_script += f"exec, coupling_knob({self.beam});\n"
        
        if ats_md:
            madx_script += "exec, full_response_ats();\n"

        return madx_script

    def get_update_correction_script(self, outpath: Path, corr_file: Path) -> str:
        madx_script = self.get_base_madx_script()
        madx_script += (
            f"call, file = '{str(corr_file)}';\n"
            f"exec, do_twiss_elements(LHCB{self.beam}, '{str(outpath)}', {self.dpp});\n"
        )
        return madx_script

    def _uses_ats_knobs(self) -> bool:
        """
        Returns wether the ATS knobs and macros should be used, based on the instance's year.
        If the **--ats** flag was explicitely provided then the returned value will be `True`.
        """
        try:
            if self.ats:
                return True
            return 2018 <= int(self.year) <= 2021  # self.year is always a string
        except ValueError:  # if a "hllhc1.x" version is given
            return False

    def _uses_run3_macros(self) -> bool:
        """Returns wether the Run 3 macros should be called, based on the instance's year."""
        try:
            return int(self.year) >= 2022  # self.year is always a string
        except ValueError:  # if a "hllhc1.x" year is given
            return False

# General functions ##########################################################


def _get_call_main_for_year(year: str) -> str:
    call_main = f"call, file = '{_get_file_for_year(year, 'main.seq')}';\n"
    if year == "2012":
        call_main += f"call, file = '{LHC_DIR / '2012' / 'install_additional_elements.madx'}';\n"
    if year == "hllhc1.3":
        call_main += f"call, file = '{LHC_DIR / 'hllhc1.3' / 'main_update.seq'}';\n"
    return call_main


def _get_file_for_year(year: str, filename: str) -> Path:
    return LHC_DIR / year / filename


def _merge_jsons(*files) -> dict:
    full_dict = {}
    for json_file in files:
        with open(json_file, "r") as json_data:
            json_dict = json.load(json_data)
            for key in json_dict.keys():
                full_dict[key] = json_dict[key]
    return full_dict


def _flatten_list(my_list: List) -> List:
    return [item for sublist in my_list for item in sublist]


def _remove_dups_keep_order(my_list: List) -> List:
    return list(OrderedDict.fromkeys(my_list))


def _list_intersect_keep_order(primary_list: List, secondary_list: List) -> List:
    return [elem for elem in primary_list if elem in secondary_list]


def _is_one_of_in(bpms_to_find: Sequence[str], bpms: Sequence[str]):
    found_bpms = [bpm for bpm in bpms_to_find if bpm in bpms]
    if len(found_bpms):
        return list(bpms).index(found_bpms[0]), found_bpms[0]
    raise KeyError


class _LhcSegmentMixin:
    def __init__(self):
        self._start = None
        self._end = None

    def get_segment_vars(self, classes=None):
        return self.get_variables(frm=self.start.s, to=self.end.s, classes=classes)

    def verify_object(self) -> None:
        try:
            self.beam
        except AttributeError:
            raise AcceleratorDefinitionError(
                "The accelerator definition is incomplete, beam "
                "has to be specified (--beam option missing?)."
            )
        if self.modifiers is None or not len(self.modifiers):
            raise AcceleratorDefinitionError(
                "The accelerator definition is incomplete, no modifiers could be found."
            )
        if self.xing is None:
            raise AcceleratorDefinitionError("Crossing on or off not set.")
        if self.label is None:
            raise AcceleratorDefinitionError("Segment label not set.")
        if self.start is None:
            raise AcceleratorDefinitionError("Segment start not set.")
        if self.end is None:
            raise AcceleratorDefinitionError("Segment end not set.")



"""
Kick
----

This module contains kick functionality of ``optics_measurements``.
It provides functions to compute kick actions.
"""
from collections import OrderedDict
from contextlib import suppress
from os.path import join

import numpy as np
import pandas as pd
import tfs

from omc3.definitions.constants import PLANE_TO_NUM
from omc3.model.accelerators.accelerator import AccElementTypes
from omc3.optics_measurements.constants import (ACTION, AMPLITUDE, BETA, DPP,
                                                DPPAMP, ERR, EXT, KICK_NAME,
                                                NAT_TUNE, PEAK2PEAK,
                                                RES,
                                                RESCALE_FACTOR, RMS,
                                                SQRT_ACTION, TIME, TUNE, S)


def calculate(measure_input, input_files, scale, header_dict, plane):
    """

    Args:
        measure_input: `OpticsInput` object.
        input_files: Stores the input files tfs.
        scale: measured beta functions.
        header_dict: `OrderedDict` containing information about the analysis.
        plane: marking the horizontal or vertical plane, **X** or **Y**.

    Returns:
        `TfsDataFrame` containing actions and their errors.
    """
    try:
        kick_frame = _getkick(measure_input, input_files, plane)
    except IndexError:  # occurs if either no x or no y files exist
        return pd.DataFrame
    kick_frame = _rescale_actions(kick_frame, scale, plane)
    header = _get_header(header_dict, plane, scale)
    tfs.write(join(measure_input.outputdir, f"{KICK_NAME}{plane.lower()}{EXT}"), kick_frame, header)
    return kick_frame.loc[:, [f"{SQRT_ACTION}{plane}", f"{ERR}{SQRT_ACTION}{plane}"]].to_numpy()


def _get_header(header_dict, plane, scale):
    header = header_dict.copy()
    header[f"{RESCALE_FACTOR}{plane}"] = scale
    return header


def _rescale_actions(df, scaling_factor, plane):
    for col in (f"{SQRT_ACTION}{plane}", f"{ERR}{SQRT_ACTION}{plane}", f"{ACTION}{plane}", f"{ERR}{ACTION}{plane}"):
        df[f"{col}{RES}"] = df.loc[:, col].to_numpy() * scaling_factor
    return df


def _getkick(measure_input, files, plane):
    load_columns, calc_columns, column_types = _get_column_mapping(plane)
    kick_frame = pd.DataFrame(data=0.,
                              index=range(len(files[plane])),
                              columns=list(load_columns.keys()) + calc_columns)

    for i, df in enumerate(files[plane]):
        # load data directly from file
        for col, src in load_columns.items():
            with suppress(KeyError):
                kick_frame.loc[i, col] = df[src]

        # calculate data from measurement
        kick_frame.loc[i, calc_columns] = _gen_kick_calc(measure_input, df, plane)
    kick_frame = kick_frame.astype(column_types)
    return kick_frame


def _gen_kick_calc(meas_input, lin, plane):
    """
    Takes either PK2PK/2 for kicker excitation or AMP for AC-dipole excitation
    """
    frame = pd.merge(_get_model_arc_betas(meas_input, plane), lin.loc[:, [f"{AMPLITUDE}{plane}", PEAK2PEAK]],
                     how='inner', left_index=True, right_index=True)
    amps = (frame.loc[:, f"{AMPLITUDE}{plane}"].to_numpy() if meas_input.accelerator.excitation
            else frame.loc[:, PEAK2PEAK].to_numpy() / 2.0)
    meansqrt2j = amps / np.sqrt(frame.loc[:, f"{BETA}{plane}"].to_numpy())
    mean2j = np.square(amps) / frame.loc[:, f"{BETA}{plane}"].to_numpy()
    return np.array([np.mean(meansqrt2j), np.std(meansqrt2j), np.mean(mean2j), np.std(mean2j)])


def _get_model_arc_betas(measure_input, plane):
    accel = measure_input.accelerator
    model = (accel.model_driven if accel.excitation else accel.model)
    return model.loc[:, [S, f"{BETA}{plane}"]].loc[
           accel.get_element_types_mask(model.index, [AccElementTypes.ARC_BPMS]), :]


def _get_column_mapping(plane):
    plane_number = PLANE_TO_NUM[plane]
    load_columns = OrderedDict([
        (TIME,                      "TIME"),
        (DPP,                       DPP),
        (DPPAMP,                    DPPAMP),
        (f"{TUNE}{plane}",          f"{TUNE}{plane_number}"),
        (f"{ERR}{TUNE}{plane}",     f"{TUNE}{plane_number}{RMS}"),
        (f"{NAT_TUNE}{plane}",      f"{NAT_TUNE}{plane_number}"),
        (f"{ERR}{NAT_TUNE}{plane}", f"{NAT_TUNE}{plane_number}{RMS}"),
    ])
    calc_columns = [f"{SQRT_ACTION}{plane}", f"{ERR}{SQRT_ACTION}{plane}",
                    f"{ACTION}{plane}", f"{ERR}{ACTION}{plane}"]

    column_types = {TIME: str}
    column_types.update({k: float for k in list(load_columns.keys())[1:] + calc_columns})
    return load_columns, calc_columns, column_types



"""
Beta from Amplitude
--------------------

This module contains some of the beta calculation related functionality of ``optics_measurements``.
It provides functions to calculate beta functions from amplitude data.
"""
from os.path import join

import numpy as np
import pandas as pd
import tfs

from omc3.optics_measurements.constants import (AMP_BETA_NAME, DELTA, ERR, EXT,
                                                MDL, RES)
from omc3.optics_measurements.toolbox import df_ratio, df_rel_diff


def calculate(meas_input, input_files, tune_dict, beta_phase, header_dict, plane):
    """
    Calculates beta and fills the following `TfsFiles`: ``f"{AMP_BETA_NAME}{plane.lower()}{EXT}"``

    Args:
        meas_input: `OpticsInput` object.
        input_files: `InputFiles` object contains measurement files.
        tune_dict: `TuneDict` contains measured tunes.
        beta_phase: contains beta functions from measured from phase.
        header_dict: dictionary of header items common for all output files.
        plane: marking the horizontal or vertical plane, **X** or **Y**.

    Returns:
    """
    beta_amp = beta_from_amplitude(meas_input, input_files, plane, tune_dict)
    x_ratio = phase_to_amp_ratio(meas_input, beta_phase, beta_amp, plane)
    beta_amp = add_rescaled_beta_columns(beta_amp, x_ratio, plane)
    header_d = _get_header(header_dict, np.std(beta_amp.loc[:, f"{DELTA}BET{plane}"].to_numpy()), x_ratio)
    tfs.write(join(meas_input.outputdir, f"{AMP_BETA_NAME}{plane.lower()}{EXT}"), beta_amp, header_d, save_index='NAME')
    return x_ratio


def phase_to_amp_ratio(measure_input, beta_phase, beta_amp, plane):
    ratio = pd.merge(beta_phase.loc[:, [f"BET{plane}"]], beta_amp.loc[:, [f"BET{plane}"]],
                     how='inner', left_index=True, right_index=True, suffixes=("ph", "amp"))
    ph_over_amp = df_ratio(ratio, f"BET{plane}ph", f"BET{plane}amp")
    mask = (np.array(0.1 < np.abs(ph_over_amp)) & np.array(np.abs(ph_over_amp) < 10.0) &
            np.array(measure_input.accelerator.get_element_types_mask(ratio.index, ["arc_bpm"])))
    x_ratio = np.mean(ph_over_amp[mask])
    return x_ratio


def add_rescaled_beta_columns(df, ratio, plane):
    df[f"BET{plane}{RES}"] = df.loc[:, f"BET{plane}"].to_numpy() * ratio
    df[f"{ERR}BET{plane}{RES}"] = df.loc[:, f"{ERR}BET{plane}"].to_numpy() * ratio
    return df


def beta_from_amplitude(meas_input, input_files, plane, tunes):
    df = pd.DataFrame(meas_input.accelerator.model).loc[:, ["S", f"MU{plane}", f"BET{plane}"]]
    df.rename(columns={f"MU{plane}": f"MU{plane}{MDL}",
                       f"BET{plane}": f"BET{plane}{MDL}"}, inplace=True)
    dpp_value = meas_input.dpp if "dpp" in meas_input.keys() else 0
    df = pd.merge(df, input_files.joined_frame(plane, [f"AMP{plane}", f"MU{plane}"], dpp_value=dpp_value),
                  how='inner', left_index=True, right_index=True)
    df['COUNT'] = len(input_files.get_columns(df, f"AMP{plane}"))

    if meas_input.compensation == "model":
        df = _compensate_by_model(input_files, meas_input, df, plane)
    if meas_input.compensation == "equation":
        df = _compensate_by_equation(input_files, meas_input, df, plane, tunes)

    amps_squared = np.square(input_files.get_data(df, f"AMP{plane}"))
    mask = meas_input.accelerator.get_element_types_mask(df.index, ["arc_bpm"])
    actions = amps_squared / df.loc[:, f"BET{plane}{MDL}"].to_numpy()[:, np.newaxis]
    betas = amps_squared / np.mean(actions[mask], axis=0, keepdims=True)
    df[f"BET{plane}"] = np.mean(betas, axis=1)
    df[f"{ERR}BET{plane}"] = np.std(betas, axis=1)
    df[f"{DELTA}BET{plane}"] = df_rel_diff(df, f"BET{plane}", f"BET{plane}{MDL}")
    df[f"{ERR}{DELTA}BET{plane}"] = df_ratio(df, f"{ERR}BET{plane}", f"BET{plane}{MDL}")
    return df.loc[:, ['S', 'COUNT', f"BET{plane}", f"{ERR}BET{plane}", f"BET{plane}{MDL}",
                      f"MU{plane}{MDL}", f"{DELTA}BET{plane}", f"{ERR}{DELTA}BET{plane}"]]


def _compensate_by_equation(input_files, meas_input, df, plane, tunes):
    phases_meas = input_files.get_data(df, f"MU{plane}") * meas_input.accelerator.beam_direction
    driven_tune, free_tune, ac2bpmac = tunes[plane]["Q"], tunes[plane]["QF"], tunes[plane]["ac2bpm"]
    k_bpmac = ac2bpmac[2]
    phase_corr = ac2bpmac[1] - phases_meas[k_bpmac] + (0.5 * driven_tune)
    phases_meas = phases_meas + phase_corr[np.newaxis, :]
    r = tunes.get_lambda(plane)
    phases_meas[k_bpmac:, :] = phases_meas[k_bpmac:, :] - driven_tune
    amp_compensation = np.sqrt((1 + r ** 2 + 2 * r * np.cos(4 * np.pi * phases_meas)) / (1 - r ** 2))
    df[input_files.get_columns(df, f"AMP{plane}")] = input_files.get_data(df, f"AMP{plane}") * amp_compensation
    return df


def _compensate_by_model(input_files, meas_input, df, plane):
    df = pd.merge(df, pd.DataFrame(meas_input.accelerator.model_driven.loc[:, [f"BET{plane}"]]
                                   .rename(columns={f"BET{plane}": f"BET{plane}comp"})),
                  how='inner', left_index=True, right_index=True)
    amp_compensation = np.sqrt(df_ratio(df, f"BET{plane}{MDL}", f"BET{plane}comp"))
    df[input_files.get_columns(df, f"AMP{plane}")] = (input_files.get_data(df, f"AMP{plane}")
                                                      * amp_compensation[:, np.newaxis])
    return df


def _get_header(header_dict, rmsbbeat, scaling_factor):
    header = header_dict.copy()
    header['RMSbetabeat'] = rmsbbeat
    header['RescalingFactor'] = scaling_factor
    return header



"""
Kicker
------

This module contains phase correction functionality of ``harpy``.
It provides tools to correct phases of a main spectral line in a case where damped (exponentially
decaying) oscillations are analysed.
"""
import numpy as np

from omc3.utils import logging_tools
from omc3.harpy.constants import COL_TUNE, COL_MU
from omc3.definitions.constants import PI2

LOGGER = logging_tools.getLogger(__name__)


def phase_correction(bpm_data_orig, lin_frame, plane):
    """
    Corrects phase of main spectral line assuming exponentially decaying oscillations.

    Args:
        bpm_data_orig: matrix of original `TbtData`.
        lin_frame: DataFrame in which to correct the results.
        plane: marking the horizontal or vertical plane, **X** or **Y**.

    Returns:
        A `DataFrame` with corrected phases.
    """
    bpm_data = bpm_data_orig.loc[lin_frame.index, :]
    damp, dstd = _get_damping(bpm_data)
    LOGGER.debug(f"Damping factor X: {damp:2.2e} +- {dstd:2.2e}")
    int_range = np.arange(0.0, bpm_data.shape[1])
    amp = lin_frame.loc[:, 'PK2PK'].to_numpy() / 2
    tune = lin_frame.loc[:, f"{COL_TUNE}{plane}"].to_numpy() * PI2
    phase = lin_frame.loc[:, f"{COL_MU}{plane}"].to_numpy() * PI2
    damp_range = damp * int_range
    phase_range = np.outer(tune, int_range - bpm_data.shape[1] / 2) + np.outer(phase, np.ones(bpm_data.shape[1]))

    e1 = np.sum(np.exp(2 * damp_range) * np.sin(2 * phase_range), axis=1) * amp/2
    e2 = np.sum(bpm_data * np.exp(damp_range) * np.sin(phase_range), axis=1)
    e3 = np.sum(bpm_data * np.exp(damp_range) * np.cos(phase_range), axis=1)
    e4 = np.sum(np.exp(2 * damp_range) * np.cos(2 * phase_range), axis=1) * amp
    cor = (e1 - e2) / ((e3 - e4) * PI2)
    lin_frame[f"{COL_MU}{plane}"] = lin_frame.loc[:, f"{COL_MU}{plane}"].to_numpy() + cor
    return lin_frame


def _get_damping(bpm_data):
    coefs = np.polyfit(np.arange(bpm_data.shape[1]),
                       np.maximum.accumulate(np.log(np.abs(bpm_data[::-1]))).T, 1)
    return np.mean(coefs[0, :]), np.std(coefs[0, :])



"""
Mock
----

Provides mock functionality for packages necessitating the ``CERN GPN`` and only installable from the
``acc-py`` package index, such as ``pytimber``, ``pjlsa`` etc.

.. code-block:: python

    from omc3.utils.mock import cern_network_import
    pytimber = cern_network_import("pytimber")
    db = pytimber.LoggingDB(source="nxcals")  # will raise if pytimber not installed
"""
import importlib


class CERNNetworkMockPackage:
    """
    Mock class to raise an error if the desired package functionality is called when the package is not
    actually installed. Designed for packages installable only from inside the CERN network,
    that are declared as ``cern`` extra. See module documentation.
    """

    def __init__(self, name: str):
        self.name = name

    def __getattr__(self, item):
        raise ImportError(
            f"The '{self.name}' package does not seem to be installed but is needed for this function. "
            "Install it with the 'cern' extra dependency, which requires to be on the CERN network and to "
            "install from the acc-py package index. Refer to the documentation for more information."
        )


def cern_network_import(package: str):
    """
    Convenience function to try and import packages only available (and installable) on the CERN network.
    If installed, the module is returned, otherwise a mock class is returned, which will raise an
    insightful ``ImportError`` on attempted use.

    Args:
        package (str): name of the package to try and import.
    """
    try:
        return importlib.import_module(package)
    except ImportError:
        return CERNNetworkMockPackage(package)



"""
Amplitude Detuning Analysis
---------------------------

Entrypoint for amplitude detuning analysis.

This module provides functionality to run amplitude detuning analysis with additionally getting
BBQ data from timber, averaging and filtering this data and subtracting it from the measurement
data.

Furthermore, the orthogonal distance regression is utilized to get a linear or quadratic fit from
the measurements.


**Arguments:**

*--Required--*

- **beam** *(int)*:

    Which beam to use.


- **kick** *(PathOrStr)*:

    Location of the kick files (parent folder).


- **plane** *(str)*:

    Plane of the kicks. 'X' or 'Y' or 'XY'.

    choices: ``['X', 'Y', 'XY']``


*--Optional--*

- **bbq_filtering_method** *(str)*:

    Filtering method for the bbq to use. 'cut' cuts around a given tune,
    'minmax' lets you specify the limits and 'outliers' uses the outlier
    filtering from utils.

    choices: ``['cut', 'minmax', 'outliers']``

    default: ``outliers``


- **bbq_in** *(UnionPathStrInt)*:

    Fill number of desired data to extract from timber or path to presaved
    bbq-tfs-file. Use the string 'kick' to use the timestamps in the
    kickfile for timber extraction. Not giving this parameter skips bbq
    compensation.


- **detuning_order** *(int)*:

    Order of the detuning as int. Basically just the order of the applied
    fit.

    default: ``1``


- **fine_cut** *(float)*:

    Cut, i.e. tolerance, of the tune (fine cleaning for 'minmax' or
    'cut').


- **fine_window** *(int)*:

    Length of the moving average window, i.e the number of data points
    (fine cleaning for 'minmax' or 'cut').


- **label** *(str)*:

    Label to identify this run.


- **outlier_limit** *(float)*:

    Limit, i.e. cut, on outliers (Method 'outliers')

    default: ``0.0002``


- **output** *(PathOrStr)*:

    Output directory for the modified kickfile and bbq data.


- **tune_cut** *(float)*:

    Cuts for the tune. For BBQ cleaning (Method 'cut').


- **tunes** *(float)*:

    Tunes for BBQ cleaning (Method 'cut').


- **tunes_minmax** *(float)*:

    Tunes minima and maxima in the order x_min, x_max, y_min, y_max. For
    BBQ cleaning (Method 'minmax').


- **window_length** *(int)*:

    Length of the moving average window. (# data points)

    default: ``20``


"""
import os
from collections import OrderedDict
from datetime import timedelta
from pathlib import Path
from typing import List, Sequence, Tuple, Dict, Any, Union

import numpy as np
import tfs
from generic_parser import DotDict
from generic_parser.entrypoint_parser import EntryPointParameters, entrypoint, save_options_to_config
from numpy.typing import ArrayLike
from tfs.frame import TfsDataFrame

from omc3.definitions import formats
from omc3.definitions.constants import PLANES
from omc3.tune_analysis import fitting_tools, kick_file_modifiers, timber_extract
from omc3.tune_analysis.bbq_tools import OutlierFilterOpt, MinMaxFilterOpt
from omc3.tune_analysis.constants import (
    get_bbq_col,
    get_bbq_out_name,
    get_kick_out_name,
    get_mav_col,
    get_timber_bbq_key, INPUT_KICK, INPUT_PREVIOUS, CORRECTED,
)
from omc3.tune_analysis.kick_file_modifiers import (
    read_timed_dataframe,
    read_two_kick_files_from_folder,
    write_timed_dataframe, AmpDetData,
)
from omc3.utils.iotools import PathOrStr, UnionPathStrInt, save_config
from omc3.utils.logging_tools import get_logger, list2str
from omc3.utils.time_tools import CERNDatetime

# Globals ----------------------------------------------------------------------

DTIME: int = 120  # extra seconds to add to kick times window when extracting from timber

LOG = get_logger(__name__)

FILTER_OPTS = dict(
    cut=("window_length", "tunes", "tune_cut"),  # "fine_window", "fine_cut"
    minmax=("window_length", "tunes_minmax", ),  # "fine_window", "fine_cut"
    outliers=("window_length", "outlier_limit"),
)


# Get Parameters ---------------------------------------------------------------


def _get_params():
    return EntryPointParameters(
        beam=dict(
            help="Which beam to use.",
            required=True,
            type=int,
        ),
        kick=dict(
            help="Location of the kick files (parent folder).",
            type=PathOrStr,
            required=True,
        ),
        plane=dict(
            help="Plane of the kicks. 'X' or 'Y' or 'XY'.",
            required=True,
            choices=list(PLANES) + ["".join(PLANES)],
            type=str,
        ),
        label=dict(
            help="Label to identify this run.",
            type=str,
        ),
        bbq_in=dict(
            help="Fill number of desired data to extract from timber or path to presaved bbq-tfs-file. "
            f"Use the string '{INPUT_KICK}' to use the timestamps in the kickfile for timber extraction. "
            f"Use the string '{INPUT_PREVIOUS}' to look for the modified ampdet kick-file from a previous run. "
            "Not giving this parameter skips bbq compensation.",
            type=UnionPathStrInt
        ),
        detuning_order=dict(
            help="Order of the detuning as int. Basically just the order of the applied fit.",
            type=int,
            default=1,
        ),
        output=dict(
            help="Output directory for the modified kickfile and bbq data.",
            type=PathOrStr,
        ),
        window_length=dict(
            help="Length of the moving average window. (# data points)",
            type=int,
            default=20,
        ),
        bbq_filtering_method=dict(
            help="Filtering method for the bbq to use. 'cut' cuts around a given tune, 'minmax' lets you "
            "specify the limits and 'outliers' uses the outlier filtering from utils.",
            type=str,
            choices=list(FILTER_OPTS.keys()),
            default="outliers",
        ),
        # Filtering method outliers
        outlier_limit=dict(
            help="Limit, i.e. cut, on outliers (Method 'outliers')",
            type=float,
            default=2e-4,
        ),
        # Filtering method tune-cut
        tunes=dict(
            help="Tunes for BBQ cleaning (Method 'cut').",
            type=float,
            nargs=2,
        ),
        tune_cut=dict(
            help="Cuts for the tune. For BBQ cleaning (Method 'cut').",
            type=float,
        ),
        # Filtering method tune-minmax
        tunes_minmax=dict(
            help="Tunes minima and maxima in the order x_min, x_max, y_min, y_max. "
            "For BBQ cleaning (Method 'minmax').",
            type=float,
            nargs=4,
        ),
        # Fine Cleaning
        fine_window=dict(
            help="Length of the moving average window, i.e the number of data points "
            "(fine cleaning for 'minmax' or 'cut').",
            type=int,
        ),
        fine_cut=dict(
            help="Cut, i.e. tolerance, of the tune (fine cleaning for 'minmax' or 'cut').",
            type=float,
        ),
    )


# Main -------------------------------------------------------------------------


@entrypoint(_get_params(), strict=True)
def analyse_with_bbq_corrections(opt: DotDict) -> Tuple[TfsDataFrame, TfsDataFrame]:
    """
    Create amplitude detuning analysis with BBQ correction from timber data.

    Returns:
        The amplitude detuning analysis results as a TfsDataFrame and the BBQ data as a TfsDataFrame.
    """
    LOG.info("Starting Amplitude Detuning Analysis")
    _save_options(opt)

    opt, filter_opt = _check_analyse_opt(opt)
    kick_df, bbq_df = get_kick_and_bbq_df(kick=opt.kick, bbq_in=opt.bbq_in,
                                          beam=opt.beam, filter_opt=filter_opt, output=opt.output)

    kick_plane = opt.plane

    for corrected in [False] + _should_do_corrected(kick_df, opt.bbq_in):
        if kick_plane in PLANES:
            kick_df = single_action_analysis(kick_df, kick_plane, opt.detuning_order, corrected)
        else:
            kick_df = double_action_analysis(kick_df, opt.detuning_order, corrected)

    if opt.output:
        LOG.info(f"Writing kick data to file in directory '{opt.output.absolute()}'")
        opt.output.mkdir(parents=True, exist_ok=True)
        write_timed_dataframe(opt.output / get_kick_out_name(), kick_df)

    return kick_df, bbq_df


def get_kick_and_bbq_df(kick: Union[Path, str], bbq_in: Union[Path, str],
                        beam: int = None, filter_opt = None,
                        output: Path = None
                        ) -> Tuple[tfs.TfsDataFrame, tfs.TfsDataFrame]:
    """Load the input data."""
    bbq_df = None
    if bbq_in is not None and bbq_in == INPUT_PREVIOUS:
        LOG.debug("Getting data from previous ampdet kick file")
        kick_df = read_timed_dataframe(Path(kick) / get_kick_out_name())
        kick_df.headers = {k: v for k, v in kick_df.headers.items() if not k.startswith("ODR_")}
    else:
        LOG.debug("Getting data from kick files")
        kick_df = read_two_kick_files_from_folder(kick)

        if bbq_in is not None:
            bbq_df = _get_bbq_data(beam, bbq_in, kick_df)

            LOG.debug("Adding moving average data to kick data")
            kick_df, bbq_df = kick_file_modifiers.add_moving_average(kick_df, bbq_df, filter_opt)

            LOG.debug("Adding corrected natural tunes and stdev to kick data")
            kick_df = kick_file_modifiers.add_corrected_natural_tunes(kick_df)

            if output:
                LOG.info(f"Writing BBQ data to file in directory '{output.absolute()}'")
                try:
                    window = filter_opt.window
                except AttributeError:
                    window = filter_opt[0].window

                x_interval = get_approx_bbq_interval(bbq_df, kick_df.index, window)
                write_timed_dataframe(output / get_bbq_out_name(), bbq_df.loc[x_interval[0]: x_interval[1]])
    return kick_df, bbq_df


def single_action_analysis(kick_df: tfs.TfsDataFrame, kick_plane: str, detuning_order: int = 1, corrected: bool = False
                           ) -> tfs.TfsDataFrame:
    """Performs the fit one action and tune pane at a time."""
    LOG.info(f"Performing amplitude detuning ODR for single-plane kicks in {kick_plane}.")
    for tune_plane in PLANES:
        LOG.debug("Getting ampdet data")
        data = kick_file_modifiers.get_ampdet_data(
            kickac_df=kick_df,
            action_plane=kick_plane,
            tune_plane=tune_plane,
            corrected=corrected
        )

        LOG.debug("Fitting ODR to kick data")
        odr_fit = fitting_tools.do_odr(
            x=data.action,
            y=data.tune,
            xerr=data.action_err,
            yerr=data.tune_err,
            order=detuning_order,
        )

        kick_df = kick_file_modifiers.add_odr(
            kickac_df=kick_df,
            odr_fit=odr_fit,
            action_plane=kick_plane,
            tune_plane=tune_plane,
            corrected=corrected
        )
    return kick_df


def double_action_analysis(kick_df: tfs.TfsDataFrame, detuning_order: int = 1, corrected: bool = False
                           ) -> tfs.TfsDataFrame:
    """Performs the full 2D/4D fitting of the data."""
    if detuning_order > 1:
        raise NotImplementedError(f"2D Analysis for detuning order {detuning_order:d} is not implemented "
                                  f"(only first order so far).")
    LOG.info("Performing amplitude detuning ODR for diagonal kicks.")
    data = {}

    # get all action arrays and all tune arrays, unfolded below
    for plane in PLANES:
        LOG.debug(f"Getting action and tune data for plane {plane}.")
        data[plane] = kick_file_modifiers.get_ampdet_data(
            kickac_df=kick_df,
            action_plane=plane,
            tune_plane=plane,
            corrected=corrected,
            dropna=False,  # so that they still have the same lengths
        )

    LOG.debug("Fitting ODR to kick data")
    odr_fits = fitting_tools.do_2d_kicks_odr(
        x=_get_ampdet_data_as_array(data, "action"),  # gets [2Jx, 2Jy]
        y=_get_ampdet_data_as_array(data, "tune"),  # gets [Qx, Qy]
        xerr=_get_ampdet_data_as_array(data, "action_err"),
        yerr=_get_ampdet_data_as_array(data, "tune_err"),
    )

    # add the fits to the kick header
    for t_plane in PLANES:
        for k_plane in PLANES:
            kick_df = kick_file_modifiers.add_odr(
                kickac_df=kick_df,
                odr_fit=odr_fits[t_plane][k_plane],
                action_plane=k_plane,
                tune_plane=t_plane,
                corrected=corrected
            )
    return kick_df


def get_approx_bbq_interval(
        bbq_df: TfsDataFrame, time_array: Sequence[CERNDatetime], window: int) -> Tuple[CERNDatetime, CERNDatetime]:
    """Get approximate start and end times for averaging, based on window length and kick interval."""
    bbq_tmp = bbq_df.dropna()

    # convert to float to use math-comparisons
    ts_bbq_index = kick_file_modifiers.get_timestamp_index(bbq_tmp.index)
    ts_kick_index = kick_file_modifiers.get_timestamp_index(time_array)
    ts_start, ts_end = min(ts_kick_index), max(ts_kick_index)

    ts_bbq_min, ts_bbq_max = min(ts_bbq_index), max(ts_bbq_index)

    if not (ts_bbq_min <= ts_start <= ts_bbq_max):
        raise ValueError("The starting time of the kicks lies outside of the given BBQ times.")

    if not (ts_bbq_min <= ts_end <= ts_bbq_max):
        raise ValueError("The end time of the kicks lies outside of the given BBQ times.")

    i_start = max(ts_bbq_index.get_indexer([ts_start], method="nearest")[0] - window, 0)
    i_end = min(ts_bbq_index.get_indexer([ts_end], method="nearest")[0] + window, len(ts_bbq_index) - 1)

    return bbq_tmp.index[i_start], bbq_tmp.index[i_end]


# Private Functions ------------------------------------------------------------


def _check_analyse_opt(opt: DotDict):
    """Perform manual checks on opt-sturcture."""
    LOG.debug("Checking Options.")

    # for label
    if opt.label is None:
        opt.label = f"Amplitude Detuning for Beam {opt.beam:d}"

    filter_opt = None
    if (opt.bbq_in is not None) and (opt.bbq_in != INPUT_PREVIOUS):
        # check if cleaning is properly specified
        method = opt.bbq_filtering_method
        if method is None:
            raise ValueError(f"Please choose a filtering method for the BBQ data from {list(FILTER_OPTS.keys())}")

        missing_params = [name for name in FILTER_OPTS[method] if opt[name] is None]
        if any(missing_params):
            raise KeyError(
                f"Missing parameters for cleaning method {method}: '{list2str(missing_params)}'"
            )

        if bool(opt.fine_cut) != bool(opt.fine_window):
            raise KeyError(
                "To activate fine cleaning, both fine cut and fine window need to be specified"
            )

        if method == "outliers":
            filter_opt = OutlierFilterOpt(
                    window=opt.window_length,
                    limit=opt.outlier_limit
               )

        elif method == "minmax":
            filter_opt = [
                MinMaxFilterOpt(
                    window=opt.window_length,
                    min=opt.tunes_minmax[2*i],
                    max=opt.tunes_minmax[2*i+1],
                    fine_window=opt.fine_window,
                    fine_cut=opt.fine_cut,
                )
                for i in range(2)
            ]
        else:
            filter_opt = [
                MinMaxFilterOpt(
                    window=opt.window_length,
                    min=opt.tunes[i] - opt.tune_cut,
                    max=opt.tunes[i] + opt.tune_cut,
                    fine_window=opt.fine_window,
                    fine_cut=opt.fine_cut,
                )
                for i in range(2)
            ]

    if opt.output is not None:
        opt.output = Path(opt.output)

    return opt, filter_opt


def _get_bbq_data(beam: int, input_: Union[Path, str, int], kick_df: TfsDataFrame) -> TfsDataFrame:
    """
    Return BBQ data from input, either file or timber fill, as a ``TfsDataFrame``.

    Note: the ``input_`` parameter is always parsed from the commandline as a string, but could be 'kick'
    or a kickfile name or an integer. All these options will be tried until one works.
    """
    try:
        fill_number = int(input_)
    except (TypeError, ValueError) as e:  # input_ is a file name or the string 'kick'
        if input_ == INPUT_KICK:
            LOG.debug("Getting timber data from kick times")
            timber_keys, bbq_cols = _get_timber_keys_and_bbq_columns(beam)
            t_start = min(kick_df.index.to_numpy())
            t_end = max(kick_df.index.to_numpy())
            t_delta = timedelta(seconds=DTIME)
            data = timber_extract.extract_between_times(
                t_start - t_delta, t_end + t_delta, keys=timber_keys, names=dict(zip(timber_keys, bbq_cols))
            )
        else:  # input_ is a file name or path
            LOG.debug(f"Getting bbq data from file '{str(input_):s}'")
            data = read_timed_dataframe(input_)
            if not len(data.index):
                raise ValueError(f"No entries in {str(input_):s}.")

    else:  # input_ is a number, assumed to be a fill number
        LOG.debug(f"Getting timber data from fill '{input_:d}'")
        timber_keys, bbq_cols = _get_timber_keys_and_bbq_columns(beam)
        data = timber_extract.lhc_fill_to_tfs(
            fill_number, keys=timber_keys, names=dict(zip(timber_keys, bbq_cols))
        )
    return data


def _get_timber_keys_and_bbq_columns(beam: int) -> Tuple[List[str], List[str]]:
    keys = [get_timber_bbq_key(plane, beam) for plane in PLANES]
    cols = [get_bbq_col(plane) for plane in PLANES]
    return keys, cols


def _should_do_corrected(kick_df, bbq_in) -> List:
    if bbq_in is None:
        return []
    if bbq_in == INPUT_PREVIOUS and not any(CORRECTED in col for col in kick_df.columns):
        return []
    return [True]


def _save_options(opt: DotDict) -> None:
    if opt.output:
        save_config(Path(opt.output), opt, __file__)


def _get_ampdet_data_as_array(data: Dict[Any, AmpDetData], column: str) -> ArrayLike:
    """ Returns a matrix with number of rows as entries in data,
    each containing the values from the given column of the AmpDetData.
    e.g. [[Jx0, Jx1, Jx2, ....]
          [Jy0, Jy1, Jy2, ....]]
    """
    return np.vstack([getattr(d, column) for d in data.values()])


# Script Mode ##################################################################


if __name__ == "__main__":
    analyse_with_bbq_corrections()



"""
Plotting Utilities: Lines
-------------------------

Line-plotting related functionality.
"""
import matplotlib
import matplotlib.transforms as mtrans
import numpy as np
from matplotlib import transforms
from matplotlib.markers import MarkerStyle
from matplotlib.patches import PathPatch

VERTICAL_LINES_TEXT_LOCATIONS = {
    'bottom': dict(y=-0.01, va='top', ha='center'),
    'top': dict(y=1.01, va='bottom', ha='center'),
    'line bottom': dict(y=0.01, va='bottom', ha='right', rotation=90),
    'line top': dict(y=0.99, va='top', ha='right', rotation=90),
}
VERTICAL_LINES_ALPHA = 0.5


class MarkerList(object):
    """Create a list of predefined markers."""
    # markers = ["s", "o", ">", "D", "v", "*", "h", "^", "p", "X", "<", "P"]  # matplotlib 2.++
    markers = ["s", "o", ">", "D", "v", "*", "h", "^", "p", "<"]

    def __init__(self):
        self.idx = 0

    @classmethod
    def get_marker(cls, marker_num):
        """ Return marker of index marker_num

         Args:
             marker_num (int): return maker at this position in list (mod len(list))
        """
        return cls.markers[marker_num % len(cls.markers)]

    def get_next_marker(self):
        """ Return the next marker in the list (circularly wrapped) """
        marker = self.get_marker(self.idx)
        self.idx += 1
        return marker


def text_to_marker(text: str, center: bool = True) -> np.ndarray:
    """
    Convert the given `text` to path
    which can be used as a marker in matplotlib plots,
    including the options of setting `markerfacecolor` and `markeredgecolor`.

    Args:
        text (str): Text to use as a marker.
        center (bool): Center the path around the origin (otherwise bottom left is anchor).

    Returns:
        Array of the path vertices.
    """
    path = MarkerStyle(fr'$\mathrm{{{text}}}$').get_path()

    if center:
        # center path: remove any offsets (extend.min) and move to center (-size[0]/2)
        extend = path.get_extents()
        t = mtrans.Affine2D().translate(-extend.min[0] - extend.size[0] / 2, -extend.min[1] - extend.size[1] / 2)
        path = path.transformed(t)

    pp = PathPatch(path, transform=mtrans.IdentityTransform())
    return pp.get_path()


def plot_vertical_lines_fast(ax, x, y=(0, 1), **kwargs):
    """
    Plots vertical lines, similar to axvline, but also in 3D, all at once and with only one label.

    Args:
        ax: Axis to plot on.
        x: array of x-positions (data coordinates).
        y: tuple of y-limits (axis coordinates, default (0,1) i.e. bottom to top).
        kwargs: kwargs passed on to ax.plot.
    """
    trans = mtrans.blended_transform_factory(ax.transData, ax.transAxes)  # x is data, y is axes
    ax.plot(np.repeat(x, 3), np.tile([*y, np.nan], len(x)), transform=trans, **kwargs)


def plot_vertical_line(ax, axvline_args: dict, text: str = None, text_loc: str = None,
                       label_size: float = matplotlib.rcParams['font.size']):
    """
    Plot a vertical line into the plot, where mline is a dictionary with arguments for ``axvline``.
    Advanced capabilities include: Automatic alpha value (if not overwritten), automatic
    zorder = -1 (if not overwritten), and adding a label to the line, where the location is given
    by text_loc.
    """
    axvline_args.setdefault('alpha', VERTICAL_LINES_ALPHA)
    axvline_args.setdefault('marker', 'None')
    axvline_args.setdefault('zorder', -1)
    if 'linestyle' not in axvline_args and 'ls' not in axvline_args:
        axvline_args['linestyle'] = '--'

    label = axvline_args.get('label', None)
    if label is None:
        axvline_args['label'] = "__nolegend__"

    line = ax.axvline(**axvline_args)

    if text is not None and text_loc is not None:
        if text_loc not in VERTICAL_LINES_TEXT_LOCATIONS:
            raise ValueError(f"Unknown value '{text_loc}' for label location.")

        ax.text(x=axvline_args['x'], s=text,
                transform=transforms.blended_transform_factory(ax.transData, ax.transAxes),
                color=line.get_color(),
                fontdict={'size': label_size}, **VERTICAL_LINES_TEXT_LOCATIONS[text_loc])



"""
Constants
---------

Specific constants to be used in ``kmod``, to help with consistency.
"""
from pathlib import Path

SEQUENCES_PATH = Path(__file__).parent / 'sequences'
EXT = ".tfs"
FIT_PLOTS_NAME = 'fit_plots.pdf'

SIDES = ("L", "R")

K = "K"
TUNE = "TUNE"
ERR = "ERR"

BETA = "BET"

STAR = "STAR"
WAIST = "WAIST"
CLEANED = "CLEANED_"
PHASEADV = "PHASEADV"
AVERAGE = "AVERAGE"
RESULTS_FILE_NAME = 'results'
LSA_FILE_NAME = 'lsa_results'
INSTRUMENTS_FILE_NAME = 'beta_instrument'



"""
PS Booster
----------

Accelerator-Class for the ``PSB`` machine.


Model Creation Keyword Args:
    *--Optional--*

    - **dpp** *(float)*:

        Deltap/p to use.

        default: ``0.0``


    - **driven_excitation** *(str)*:

        Denotes driven excitation by `AC-dipole` (acd) or by `ADT` (adt)

        choices: ``('acd', 'adt')``


    - **drv_tunes** *(float)*:

        Driven tunes without integer part.


    - **energy** *(float)*:

        Energy in **Tev**.


    - **model_dir** *(str)*:

        Path to model directory; loads tunes and excitation from model!


    - **modifiers** *(str)*:

        Path to the optics file to use (modifiers file).


    - **nat_tunes** *(float)*:

        Natural tunes without integer part.


    - **ring** *(int)*:

        Ring to use.

        choices: ``(1, 2, 3, 4)``


    - **xing**:

        If True, x-ing angles will be applied to model

        action: ``store_true``
"""
import logging
from pathlib import Path

from generic_parser import EntryPoint

from omc3.model.accelerators.accelerator import (Accelerator,
                                                 AcceleratorDefinitionError, AccElementTypes, AccExcitationMode)
from omc3.model.constants import PLANE_TO_HV, MODIFIER_TAG

LOGGER = logging.getLogger(__name__)
CURRENT_DIR = Path(__file__).parent


class Psbooster(Accelerator):
    """Parent Class for Psbooster-types."""
    NAME = "psbooster"
    RE_DICT = {AccElementTypes.BPMS: r"BR\d\.BPM[^T]",
               AccElementTypes.MAGNETS: r".*",
               AccElementTypes.ARC_BPMS: r"BR\d\.BPM[^T]"
               }

    @staticmethod
    def get_parameters():
        params = super(Psbooster, Psbooster).get_parameters()
        params.add_parameter(name="ring", type=int, choices=(1, 2, 3, 4), help="Ring to use.")
        return params

    def __init__(self, *args, **kwargs):
        parser = EntryPoint(self.get_parameters(), strict=True)
        opt = parser.parse(*args, **kwargs)
        super().__init__(opt)
        self.ring = opt.ring

    @property
    def ring(self):
        if self._ring is None:
            raise AcceleratorDefinitionError("The accelerator definition is incomplete, ring "
                                             "has to be specified (--ring option missing?).")
        return self._ring

    @ring.setter
    def ring(self, value):
        if value not in (1, 2, 3, 4):
            raise AcceleratorDefinitionError("Ring parameter has to be one of (1, 2, 3, 4)")
        self._ring = value

    def verify_object(self):
        _ = self.ring
        if self.modifiers:
            raise AcceleratorDefinitionError(f"Accelerator {self.NAME} cannot handle modifiers,"
                                             f" yet modifiers were given.")

    def get_exciter_bpm(self, plane, bpms):
        if not self.excitation:
            return None
        bpms_to_find = [f"BR{self.ring}.BPM3L3", f"BR{self.ring}.BPM4L3"]
        found_bpms = [bpm for bpm in bpms_to_find if bpm in bpms]
        if not len(found_bpms):
            raise KeyError
        return (list(bpms).index(found_bpms[0]), found_bpms[0]), f"{PLANE_TO_HV[plane]}ACMAP"

    def get_base_madx_script(self, best_knowledge=False):
        if best_knowledge:
            raise NotImplementedError(f"Best knowledge model not implemented for accelerator {self.NAME}")

        use_acd = self.excitation == AccExcitationMode.ACD
        replace_dict = {
            "FILES_DIR": str(self.get_dir()),
            "USE_ACD": str(int(use_acd)),
            "RING": self.ring,
            "NAT_TUNE_X": self.nat_tunes[0],
            "NAT_TUNE_Y": self.nat_tunes[1],
            "KINETICENERGY": self.energy,
            "DRV_TUNE_X": "",
            "DRV_TUNE_Y": "",
        }
        if use_acd:
            replace_dict["DRV_TUNE_X"] = self.drv_tunes[0]
            replace_dict["DRV_TUNE_Y"] = self.drv_tunes[1]
        mask = self.get_file('base.mask').read_text()
        return mask % replace_dict


class _PsboosterSegmentMixin(object):

    def __init__(self):
        self._start = None
        self._end = None
